'''bam2stats.py - compute stats from a bam-file
===============================================

:Tags: Genomics NGS Summary BAM

Purpose
-------

This script takes a bam file as input and computes a few metrics by
iterating over the file. The metrics output are:

+------------------------+------------------------------------------+
|*Category*              |*Content*                                 |
+------------------------+------------------------------------------+
|total                   |total number of alignments in bam file    |
+------------------------+------------------------------------------+
|alignments_mapped       |alignments mapped to a chromosome (bam    |
|                        |flag)                                     |
+------------------------+------------------------------------------+
|alignments_unmapped     |alignments unmapped (bam flag)            |
+------------------------+------------------------------------------+
|qc_fail                 |alignments failing QC (bam flag)          |
+------------------------+------------------------------------------+
|mate_unmapped           |alignments in which the mate is unmapped  |
|                        |(bam flag)                                |
+------------------------+------------------------------------------+
|reverse                 |alignments in which read maps to reverse  |
|                        |strand (bam flag)                         |
+------------------------+------------------------------------------+
|mate_reverse            |alignments in which mate maps to reverse  |
|                        |strand (bam flag)                         |
+------------------------+------------------------------------------+
|proper_pair             |alignments in which both pairs have been  |
|                        |mapped properly (according to the mapper) |
|                        |(bam flag)                                |
+------------------------+------------------------------------------+
|read1                   |alignments for 1st read of pair (bam flag)|
+------------------------+------------------------------------------+
|paired                  |alignments of reads that are paired (bam  |
|                        |flag)                                     |
+------------------------+------------------------------------------+
|duplicate               |read is PCR or optical duplicate (bam     |
|                        |flag)                                     |
+------------------------+------------------------------------------+
|read2                   |alignment is for 2nd read of pair (bam    |
|                        |flag)                                     |
+------------------------+------------------------------------------+
|secondary               |alignment is not primary alignment        |
+------------------------+------------------------------------------+
|alignments_rna          |alignments mapping to regions specified in|
|                        |a .gff file                               |
+------------------------+------------------------------------------+
|alignments_no_rna       |alignments not mapping to regions in a    |
|                        |.gff file (if --ignore-masked-reads has   |
|                        |been set, otherwise equal to mapped)      |
+------------------------+------------------------------------------+
|alignments_duplicates   |number of alignments mapping to the same  |
|                        |location                                  |
+------------------------+------------------------------------------+
|alignments_unique       |number of alignments mapping to unique    |
|                        |locations                                 |
+------------------------+------------------------------------------+
|reads_total             |number of reads in file. Either given via |
|                        |--num-reads or deduc ed as the sum of     |
|                        |mappend and unmapped reads                |
+------------------------+------------------------------------------+
|reads_mapped            |number of reads mapping in file. Derived  |
|                        |from the total number o f alignments and  |
|                        |removing counts for multiple              |
|                        |matches. Requires the NH flag to be set   |
|                        |correctly.                                |
+------------------------+------------------------------------------+
|reads_unmapped          |number of reads unmapped in file. Assumes |
|                        |that there is only one                    |
|                        |entry per unmapped read.                  |
+------------------------+------------------------------------------+
|reads_missing           |number of reads missing, if number of     |
|                        |reads given by --input-rea ds. Otherwise  |
|                        |0.                                        |
+------------------------+------------------------------------------+
|reads_norna             |reads not mapping to repetetive RNA       |
|                        |regions.                                  |
+------------------------+------------------------------------------+
|pairs_total             |number of total pairs - this is the number|
|                        |of reads_total divided by two. If there   |
|                        |were no pairs, pairs_total will be 0.     |
+------------------------+------------------------------------------+
|pairs_mapped            |number of mapped pairs - this is the same |
|                        |as the number of proper pairs.            |
+------------------------+------------------------------------------+

Additionally, the script outputs histograms for the following tags and
scores.

* NM: number of mismatches in alignments.
* NH: number of hits of reads.
* mapq: mapping quality of alignments.

Supplying a fastq file
++++++++++++++++++++++

If a fastq file is supplied (``--fastq-file``), the script will
compute some additional summary statistics. However, as it builds a dictionary
of all sequences, it will also require a good  amount of memory. The additional
metrics output are:

+-----------------------------+----------------------------------------+
|*Category*                   |*Content*                               |
+-----------------------------+----------------------------------------+
|pairs_total                  |total number of pairs in input data     |
+-----------------------------+----------------------------------------+
|pairs_mapped                 |pairs in which both reads map           |
+-----------------------------+----------------------------------------+
|pairs_unmapped               |pairs in which neither read maps        |
+-----------------------------+----------------------------------------+
|pairs_proper_unique          |pairs which are proper and map uniquely.|
+-----------------------------+----------------------------------------+
|pairs_incomplete_unique      |pairs in which one of the reads maps    |
|                             |uniquely, but the other does not map.   |
+-----------------------------+----------------------------------------+
|pairs_incomplete_multimapping|pairs in which one of the reads maps    |
|                             |uniquely, but the other maps to multiple|
|                             |locations.                              |
+-----------------------------+----------------------------------------+
|pairs_proper_duplicate       |pairs which are proper and unique, but  |
|                             |marked as duplicates.                   |
+-----------------------------+----------------------------------------+
|pairs_proper_multimapping    |pairs which are proper, but map to      |
|                             |multiple locations.                     |
+-----------------------------+----------------------------------------+
|pairs_not_proper_unique      |pairs mapping uniquely, but not flagged |
|                             |as proper                               |
+-----------------------------+----------------------------------------+
|pairs_other                  |pairs not in any of the above categories|
+-----------------------------+----------------------------------------+

Note that for paired-end data, any ``\1`` or ``/2`` suffixes will be
removed from the read name in the assumption that these have been removed
in the bam file as well.

Usage
-----

Example::

   python bam2stats.py in.bam

This command will generate various statistics based on the supplied
BAM file, such as percentage reads mapped and percentage reads mapped
in pairs. The output looks like this:

+-----------------------------+------+-------+-----------------+
|category                     |counts|percent|of               |
+-----------------------------+------+-------+-----------------+
|alignments_total             |32018 |100.00 |alignments_total |
+-----------------------------+------+-------+-----------------+
|alignments_mapped            |32018 |100.00 |alignments_total |
+-----------------------------+------+-------+-----------------+
|alignments_unmapped          |0     | 0.00  |alignments_total |
+-----------------------------+------+-------+-----------------+
|alignments_qc_fail           |0     | 0.00  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_mate_unmapped     |241   | 0.75  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_reverse           |16016 |50.02  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_mate_reverse      |15893 |49.64  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_proper_pair       |30865 |96.40  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_read1             |16057 |50.15  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_paired            |32018 |100.00 |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_duplicate         |0     | 0.00  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_read2             |15961 |49.85  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_secondary         |0     | 0.00  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_rna               |68    | 0.21  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_no_rna            |31950 |99.79  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|alignments_filtered          |31950 |99.79  |alignments_mapped|
+-----------------------------+------+-------+-----------------+
|reads_total                  |34250 |100.00 |reads_total      |
+-----------------------------+------+-------+-----------------+
|reads_unmapped               |0     | 0.00  |reads_total      |
+-----------------------------+------+-------+-----------------+
|reads_mapped                 |32018 |93.48  |reads_total      |
+-----------------------------+------+-------+-----------------+
|reads_missing                |2232  | 6.52  |reads_total      |
+-----------------------------+------+-------+-----------------+
|reads_mapped_unique          |32018 |100.00 |reads_mapped     |
+-----------------------------+------+-------+-----------------+
|reads_multimapping           |0     | 0.00  |reads_mapped     |
+-----------------------------+------+-------+-----------------+
|pairs_total                  |17125 |100.00 |pairs_total      |
+-----------------------------+------+-------+-----------------+
|pairs_mapped                 |17125 |100.00 |pairs_total      |
+-----------------------------+------+-------+-----------------+
|pairs_unmapped               |0     | 0.00  |pairs_total      |
+-----------------------------+------+-------+-----------------+
|pairs_proper_unique          |14880 |86.89  |pairs_total      |
+-----------------------------+------+-------+-----------------+
|pairs_incomplete_unique      |2232  |13.03  |pairs_total      |
+-----------------------------+------+-------+-----------------+
|pairs_incomplete_multimapping|0     | 0.00  |pairs_total      |
+-----------------------------+------+-------+-----------------+
|pairs_proper_duplicate       |0     | 0.00  |pairs_total      |
+-----------------------------+------+-------+-----------------+
|pairs_proper_multimapping    |0     | 0.00  |pairs_total      |
+-----------------------------+------+-------+-----------------+
|pairs_not_proper_unique      |13    | 0.08  |pairs_total      |
+-----------------------------+------+-------+-----------------+
|pairs_other                  |0     | 0.00  |pairs_total      |
+-----------------------------+------+-------+-----------------+
|read1_total                  |17125 |100.00 |read1_total      |
+-----------------------------+------+-------+-----------------+
|read1_unmapped               |0     | 0.00  |read1_total      |
+-----------------------------+------+-------+-----------------+
|read1_mapped                 |16057 |93.76  |read1_total      |
+-----------------------------+------+-------+-----------------+
|read1_mapped_unique          |16057 |100.00 |read1_mapped     |
+-----------------------------+------+-------+-----------------+
|reads_multimapping           |0     | 0.00  |read1_mapped     |
+-----------------------------+------+-------+-----------------+
|read1_missing                |1068  | 6.65  |read1_total      |
+-----------------------------+------+-------+-----------------+
|read2_total                  |17125 |100.00 |read2_total      |
+-----------------------------+------+-------+-----------------+
|read2_unmapped               |0     | 0.00  |read2_total      |
+-----------------------------+------+-------+-----------------+
|read2_mapped                 |15961 |93.20  |read2_total      |
+-----------------------------+------+-------+-----------------+
|read2_mapped_unique          |15961 |100.00 |read2_mapped     |
+-----------------------------+------+-------+-----------------+
|reads_multimapping           |0     | 0.00  |read2_mapped     |
+-----------------------------+------+-------+-----------------+
|read2_missing                |1164  | 7.29  |read2_total      |
+-----------------------------+------+-------+-----------------+

The first column contains the caterogy, the second the number of
counts and the third a percentage. The fourth column denotes the
denomiminator that was used to compute the percentage. In the table
above, wee see that 16,057 first reads in a pair map and 15,961
second reads in pair map, resulting in 14,880 proper uniquely mapped
pairs.

Type::

   cgat bam2stats --help

for command line help.

Bam2stats can read from standard input::

   cat in.bam | python bam2stats.py -


Documentation
-------------

Reads are not counted via read name, but making use of NH and HI flags
when present.  To recap, NH is the number of reported alignments that
contain the query in the current record, while HI is the hit index and
ranges from 0 to NH-1.

Unfortunately, not all aligners follow this convention. For example,
gsnap seems to set NH to the number of reportable alignments, while
the actual number of reported alignments in the file is less. Thus, if
the HI flag is present, the maximum HI is used to correct the NH
flag. The assumption is, that the same reporting threshold has been
used for all alignments.

If no NH flag is present, it is assumed that all reads have only been
reported once.

Multi-matching counts after filtering are really guesswork. Basically,
the assumption is that filtering is consistent and will tend to remove
all alignments of a query.

Command line options
--------------------

'''

import os
import sys
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.GTF as GTF
import pysam
import CGAT.scripts._bam2stats as _bam2stats

FLAGS = {
    1: 'paired',
    2: 'proper_pair',
    4: 'unmapped',
    8: 'mate_unmapped',
    16: 'reverse',
    32: 'mate_reverse',
    64: 'read1',
    128: 'read2',
    256: 'secondary',
    512: 'qc_fail',
    1024: 'duplicate',
}


def computeMappedReadsFromAlignments(total_alignments, nh, max_hi):
    '''compute number of reads alignment from total number of alignments.
    '''
    nreads_mapped = total_alignments
    if len(nh) > 0:
        max_nh = max(nh.keys())
        if max_hi > 0:
            for x in range(2, min(max_nh + 1, max_hi)):
                nreads_mapped -= (nh[x] / x) * (x - 1)
            for x in range(max_hi, max_nh + 1):
                nreads_mapped -= (nh[x] / max_hi) * (max_hi - 1)
        else:
            for x in range(2, max(nh.keys()) + 1):
                nreads_mapped -= (nh[x] / x) * (x - 1)

    return nreads_mapped


def writeNH(outfile, nh, max_hi):
    '''output nh array, correcting for max_hi if less than nh'''

    # need to remove double counting
    # one read matching to 2 positions is only 2

    max_nh = max(nh.keys())
    if max_hi > 0:
        for x in range(1, min(max_nh + 1, max_hi)):
            if nh[x] == 0:
                continue
            outfile.write("%i\t%i\n" % (x, nh[x] / x))
        for x in range(max_hi, max_nh + 1):
            if nh[x] == 0:
                continue
            outfile.write("%i\t%i\n" % (x, nh[x] / max_hi))
    else:
        for x in range(1, max_nh + 1):
            if nh[x] == 0:
                continue
            outfile.write("%i\t%i\n" % (x, nh[x] / x))


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option(
        "-r", "--mask-bed-file", dest="filename_rna", type="string",
        metavar='GFF',
        help="gff formatted file with masking locations. The number of "
        "reads overlapping the intervals in the given file will be "
        "computed. Note that the computation currently does not take "
        "into account indels, so it is an approximate count only. "
        "[%default]")

    parser.add_option(
        "-f", "--ignore-masked-reads", dest="remove_rna", action="store_true",
        help="as well as counting reads in the file given by --mask-bed-file, "
        "also remove these reads for duplicate and match statistics. "
        "[%default]")

    parser.add_option(
        "-i", "--num-reads", dest="input_reads", type="int",
        help="the number of reads - if given, used to provide percentages "
        "[%default]")

    parser.add_option(
        "-d", "--output-details", dest="output_details", action="store_true",
        help="output per-read details into a separate file. Read names are "
        "md5/base64 encoded [%default]")

    parser.add_option(
        "-q", "--fastq-file", dest="filename_fastq",
        help="filename with sequences and quality scores. This file is only "
        "used to collect sequence identifiers. Thus, for paired end data a "
        "single file is sufficient [%default]")

    parser.set_defaults(
        filename_rna=None,
        remove_rna=False,
        input_reads=0,
        force_output=False,
        filename_fastq=None,
        output_details=False,
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv, add_output_options=True)

    if options.filename_rna:
        rna = GTF.readAndIndex(
            GTF.iterator(IOTools.openFile(options.filename_rna)))
    else:
        rna = None

    if len(args) > 0:
        pysam_in = pysam.AlignmentFile(args[0], "rb")
    elif options.stdin == sys.stdin:
        pysam_in = pysam.AlignmentFile("-", "rb")
    else:
        pysam_in = pysam.AlignmentFile(options.stdin, "rb")

    if options.output_details:
        outfile_details = E.openOutputFile("details", "w")
    else:
        outfile_details = None

    if options.filename_fastq and not os.path.exists(options.filename_fastq):
        raise IOError("file %s does not exist" % options.filename_fastq)

    (counter, flags_counts, nh_filtered, nh_all,
     nm_filtered, nm_all, mapq, mapq_all, max_hi) = \
        _bam2stats.count(pysam_in,
                         options.remove_rna,
                         rna,
                         filename_fastq=options.filename_fastq,
                         outfile_details=outfile_details)

    if max_hi > 0 and max_hi != max(nh_all.keys()):
        E.warn("max_hi(%i) is inconsistent with max_nh (%i) "
               "- counts will be corrected"
               % (max_hi, max(nh_all.keys())))

    outs = options.stdout
    outs.write("category\tcounts\tpercent\tof\n")

    def _write(outs, text, numerator, denominator, base):
        percent = IOTools.prettyPercent(numerator, denominator)
        outs.write('%s\t%i\t%s\t%s\n' % (text,
                                         numerator,
                                         percent,
                                         base))

    ###############################
    ###############################
    ###############################
    # Output alignment information
    ###############################
    nalignments_unmapped = flags_counts["unmapped"]
    nalignments_mapped = counter.alignments_input - nalignments_unmapped

    _write(outs,
           "alignments_total",
           counter.alignments_input,
           counter.alignments_input,
           "alignments_total")

    if counter.alignments_input == 0:
        E.warn("no alignments in BAM file - no further output")
        E.Stop()
        return

    _write(outs,
           "alignments_mapped",
           nalignments_mapped,
           counter.alignments_input,
           'alignments_total')
    _write(outs,
           "alignments_unmapped",
           nalignments_unmapped,
           counter.alignments_input,
           'alignments_total')

    if nalignments_mapped == 0:
        E.warn("no mapped alignments - no further output")
        E.Stop()
        return

    for flag, counts in sorted(flags_counts.items()):
        if flag == "unmapped":
            continue
        _write(outs,
               'alignments_' + flag,
               counts,
               nalignments_mapped,
               'alignments_mapped')

    if options.filename_rna:
        _write(outs,
               "alignments_rna",
               counter.alignments_rna,
               nalignments_mapped,
               'alignments_mapped')
        _write(outs,
               "alignments_no_rna",
               counter.alignments_no_rna,
               nalignments_mapped,
               'alignments_mapped')

    _write(outs,
           "alignments_filtered",
           counter.alignments_filtered,
           nalignments_mapped,
           "alignments_mapped")

    if counter.filtered == nalignments_mapped:
        normby = "alignments_mapped"
    else:
        normby = "alignments_filtered"

    if counter.filtered > 0:
        _write(outs,
               "alignments_duplicates",
               counter.alignments_duplicates,
               counter.alignments_filtered,
               normby)
        _write(outs,
               "alignments_unique",
               counter.aligmnments_filtered - counter.alignments_duplicates,
               counter.alignments_filtered,
               normby)

    ###############################
    ###############################
    ###############################
    # Output read based information
    ###############################

    # derive the number of mapped reads in file from alignment counts
    if options.filename_fastq:
        nreads_total = counter.total_read
        _write(outs,
               "reads_total",
               counter.total_read,
               nreads_total,
               'reads_total')
        _write(outs,
               "reads_unmapped",
               counter.total_read_is_unmapped,
               nreads_total,
               'reads_total')
        _write(outs,
               "reads_mapped",
               counter.total_read_is_mapped,
               nreads_total,
               'reads_total')
        _write(outs,
               "reads_missing",
               counter.total_read_is_missing,
               nreads_total,
               'reads_total')
        _write(outs,
               "reads_mapped_unique",
               counter.total_read_is_mapped_uniq,
               counter.total_read_is_mapped,
               'reads_mapped')
        _write(outs,
               "reads_multimapping",
               counter.total_read_is_mmap,
               counter.total_read_is_mapped,
               'reads_mapped')
    else:
        E.warn('inferring read counts from alignments and NH tags')
        nreads_unmapped = flags_counts["unmapped"]
        nreads_mapped = computeMappedReadsFromAlignments(nalignments_mapped,
                                                         nh_all, max_hi)

        nreads_missing = 0
        if options.input_reads:
            nreads_total = options.input_reads
            # unmapped reads in bam file?
            if nreads_unmapped:
                nreads_missing = nreads_total - nreads_unmapped - nreads_mapped
            else:
                nreads_unmapped = nreads_total - nreads_mapped

        elif nreads_unmapped:
            # if unmapped reads are in bam file, take those
            nreads_total = nreads_mapped + nreads_unmapped
        else:
            # otherwise normalize by mapped reads
            nreads_unmapped = 0
            nreads_total = nreads_mapped

        outs.write("reads_total\t%i\t%5.2f\treads_total\n" %
                   (nreads_total, 100.0))
        outs.write("reads_mapped\t%i\t%5.2f\treads_total\n" %
                   (nreads_mapped, 100.0 * nreads_mapped / nreads_total))
        outs.write("reads_unmapped\t%i\t%5.2f\treads_total\n" %
                   (nreads_unmapped, 100.0 * nreads_unmapped / nreads_total))
        outs.write("reads_missing\t%i\t%5.2f\treads_total\n" %
                   (nreads_missing, 100.0 * nreads_missing / nreads_total))

        if len(nh_all) > 1:
            outs.write("reads_unique\t%i\t%5.2f\treads_mapped\n" %
                       (nh_all[1], 100.0 * nh_all[1] / nreads_mapped))

        # compute after filtering
        # not that these are rough guesses
        if options.filename_rna:
            nreads_norna = computeMappedReadsFromAlignments(
                counter.filtered, nh_filtered, max_hi)
            _write(outs,
                   "reads_norna",
                   nreads_norna,
                   nreads_mapped,
                   "reads_mapped")
            if len(nh_filtered) > 1:
                _write(outs,
                       "reads_norna_unique",
                       nh_filtered[1],
                       nreads_norna,
                       "reads_mapped")

    pysam_in.close()

    ###############################
    ###############################
    ###############################
    # Output pair information
    ###############################
    if flags_counts["read2"] > 0:
        if options.filename_fastq:
            pairs_mapped = counter.total_pair_is_mapped

            # sanity check
            assert counter.total_pair_is_mapped == \
                (counter.total_pair_is_proper_uniq +
                 counter.total_pair_is_incomplete_uniq +
                 counter.total_pair_is_incomplete_mmap +
                 counter.total_pair_is_proper_duplicate +
                 counter.total_pair_is_proper_mmap +
                 counter.total_pair_not_proper_uniq +
                 counter.total_pair_is_other)

            outs.write("pairs_total\t%i\t%5.2f\tpairs_total\n" %
                       (counter.total_pairs,
                        100.0 * counter.total_pairs / counter.total_pairs))
            outs.write("pairs_mapped\t%i\t%5.2f\tpairs_total\n" %
                       (pairs_mapped,
                        100.0 * pairs_mapped / counter.total_pairs))
            outs.write(
                "pairs_unmapped\t%i\t%5.2f\tpairs_total\n" %
                (counter.total_pair_is_unmapped,
                 100.0 * counter.total_pair_is_unmapped / counter.total_pairs))
            outs.write(
                "pairs_proper_unique\t%i\t%5.2f\tpairs_total\n" %
                (counter.total_pair_is_proper_uniq,
                 100.0 * counter.total_pair_is_proper_uniq /
                 counter.total_pairs))
            outs.write(
                "pairs_incomplete_unique\t%i\t%5.2f\tpairs_total\n" %
                (counter.total_pair_is_incomplete_uniq,
                 100.0 * counter.total_pair_is_incomplete_uniq /
                 counter.total_pairs))
            outs.write(
                "pairs_incomplete_multimapping\t%i\t%5.2f\tpairs_total\n" %
                (counter.total_pair_is_incomplete_mmap,
                 100.0 * counter.total_pair_is_incomplete_mmap /
                 counter.total_pairs))
            outs.write(
                "pairs_proper_duplicate\t%i\t%5.2f\tpairs_total\n" %
                (counter.total_pair_is_proper_duplicate,
                 100.0 * counter.total_pair_is_proper_duplicate /
                 counter.total_pairs))
            outs.write(
                "pairs_proper_multimapping\t%i\t%5.2f\tpairs_total\n" %
                (counter.total_pair_is_proper_mmap,
                 100.0 * counter.total_pair_is_proper_mmap /
                 counter.total_pairs))
            outs.write(
                "pairs_not_proper_unique\t%i\t%5.2f\tpairs_total\n" %
                (counter.total_pair_not_proper_uniq,
                 100.0 * counter.total_pair_not_proper_uniq /
                 counter.total_pairs))
            outs.write(
                "pairs_other\t%i\t%5.2f\tpairs_total\n" %
                (counter.total_pair_is_other,
                 100.0 * counter.total_pair_is_other /
                 counter.total_pairs))

            nread1_total = counter.total_read1
            _write(outs,
                   "read1_total",
                   counter.total_read1,
                   nread1_total,
                   'read1_total')
            _write(outs,
                   "read1_unmapped",
                   counter.total_read1_is_unmapped,
                   nread1_total,
                   'read1_total')
            _write(outs,
                   "read1_mapped",
                   counter.total_read1_is_mapped,
                   nread1_total,
                   'read1_total')
            _write(outs,
                   "read1_mapped_unique",
                   counter.total_read1_is_mapped_uniq,
                   counter.total_read1_is_mapped,
                   'read1_mapped')
            _write(outs,
                   "reads_multimapping",
                   counter.total_read1_is_mmap,
                   counter.total_read1_is_mapped,
                   'read1_mapped')
            _write(outs,
                   "read1_missing",
                   counter.total_read1_is_missing,
                   counter.total_read1_is_mapped,
                   'read1_total')

            nread2_total = counter.total_read2
            _write(outs,
                   "read2_total",
                   counter.total_read2,
                   nread2_total,
                   'read2_total')
            _write(outs,
                   "read2_unmapped",
                   counter.total_read2_is_unmapped,
                   nread2_total,
                   'read2_total')
            _write(outs,
                   "read2_mapped",
                   counter.total_read2_is_mapped,
                   nread2_total,
                   'read2_total')
            _write(outs,
                   "read2_mapped_unique",
                   counter.total_read2_is_mapped_uniq,
                   counter.total_read2_is_mapped,
                   'read2_mapped')
            _write(outs,
                   "reads_multimapping",
                   counter.total_read2_is_mmap,
                   counter.total_read2_is_mapped,
                   'read2_mapped')
            _write(outs,
                   "read2_missing",
                   counter.total_read2_is_missing,
                   counter.total_read2_is_mapped,
                   'read2_total')

        else:
            # approximate counts
            pairs_total = nreads_total // 2
            pairs_mapped = flags_counts["proper_pair"] // 2
            _write(outs,
                   "pairs_total",
                   pairs_total,
                   pairs_total,
                   "pairs_total")
            _write(outs,
                   "pairs_mapped",
                   pairs_mapped,
                   pairs_total,
                   "pairs_total")
    else:
        # no paired end data
        pairs_total = pairs_mapped = 0
        outs.write("pairs_total\t%i\t%5.2f\tpairs_total\n" %
                   (pairs_total, 0.0))
        outs.write("pairs_mapped\t%i\t%5.2f\tpairs_total\n" %
                   (pairs_mapped, 0.0))

    if options.force_output or len(nm_filtered) > 0:
        outfile = E.openOutputFile("nm", "w")
        outfile.write("NM\talignments\n")
        if len(nm_filtered) > 0:
            for x in range(0, max(nm_filtered.keys()) + 1):
                outfile.write("%i\t%i\n" % (x, nm_filtered[x]))
        else:
            outfile.write("0\t%i\n" % (counter.filtered))
        outfile.close()

    if options.force_output or len(nh_all) > 1:
        outfile = E.openOutputFile("nh_all", "w")
        outfile.write("NH\treads\n")
        if len(nh_all) > 0:
            writeNH(outfile, nh_all, max_hi)
        else:
            # assume all are unique if NH flag not set
            outfile.write("1\t%i\n" % (counter.mapped_reads))
        outfile.close()

    if options.force_output or len(nh_filtered) > 1:
        outfile = E.openOutputFile("nh", "w")
        outfile.write("NH\treads\n")
        if len(nh_filtered) > 0:
            writeNH(outfile, nh_filtered, max_hi)
        else:
            # assume all are unique if NH flag not set
            outfile.write("1\t%i\n" % (counter.filtered))
        outfile.close()

    if options.force_output or len(mapq_all) > 1:
        outfile = E.openOutputFile("mapq", "w")
        outfile.write("mapq\tall_reads\tfiltered_reads\n")
        for x in range(0, max(mapq_all.keys()) + 1):
            outfile.write("%i\t%i\t%i\n" % (x, mapq_all[x], mapq[x]))
        outfile.close()

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
AString.py - strings as arrays of characters
============================================

This module provides the :class:`AString` class to efficiently
represent long, chromosomal nucleotide sequences in memory.

Reference
---------

'''
from array import array


import sys

IS_PY3 = sys.version_info.major >= 3


class AString(array):
    """implementation of a string as an array.

    This class conserves memory as it uses only 1 byte per letter,
    while python strings use the machine word size for a letter.

    It adds a subset of the python string class such as upper() and
    lower() for convenience. Slicing and printing return strings.

    The :class:`AString` can be constructed by any iterable that is
    accepted by the constructor of :py:class:`array.array`.

    """

    def __new__(cls, *args):
        if IS_PY3:
            return array.__new__(cls, "b", *[x.encode("ascii") for x in args])
        else:
            return array.__new__(cls, "b", *args)

    def upper(self):
        """return upper case version."""
        if IS_PY3:
            return AString(str(self).upper().encode('ascii'))
        else:
            return AString(str(self).upper())

    def lower(self):
        """return lower case version."""
        if IS_PY3:
            return AString(str(self).lower().encode('ascii'))
        else:
            return AString(str(self).lower())

    def __getitem__(self, *args):
        """return slice as a string."""

        if IS_PY3:
            return array.__getitem__(self, *args).tostring().decode("ascii")
        else:
            return array.__getitem__(self, *args).tostring()

    def __setitem__(self, start, end, sub):
        """set slice start:end from a string sub."""
        return array.__setitem__(self,
                                 start, end,
                                 array("b", sub))

    def __str__(self):
        if IS_PY3:
            return self.tostring().decode("ascii")
        else:
            return self.tostring()



'''
CSV2DB.py - utilities for uploading a table to database
=======================================================

:Tags: Python

Purpose
-------

create a table from a csv separated file and load data into it.

This module supports backends for postgres and sqlite3. Column types are
auto-detected.

.. todo::

   Use file import where appropriate to speed up loading. Currently, this is
   not always the case.

Usage
-----

Documentation
-------------

Code
----

'''
import os
import csv
import string
import re
import time
import tempfile

from CGAT import Experiment as E
from CGAT import CSV as CSV
from CGAT import IOTools as IOTools
import sqlite3


def executewait(dbhandle, statement, error,
                retry=False,
                wait=5,
                args=()):
    '''execute sql statement.

    Retry on error, if retry is True.
    Returns a cursor object.
    '''

    cc = dbhandle.cursor()
    i = 20
    while i > 0:
        try:
            cc.execute(statement, args)
            return cc
        except sqlite3.OperationalError as e:
            E.warn("import failed: msg=%s, statement=\n  %s" %
                   (str(e), statement))
        # TODO: check for database locked msg
            if not retry:
                raise e
            if not re.search("locked", str(e)):
                raise e
            time.sleep(wait)
            i -= 1
            continue
        break
    raise sqlite3.OperationalError("Database locked and too many retries")


def quoteRow(row, take,
             map_column2type,
             missing_values,
             null="NULL",
             string_value="%s"):
    """return a dictionary with properly quoted values."""

    # set empty values for int/float to NULL
    d = {}
    for t in take:
        v = row[t]
        if v == "":
            d[t] = null
        elif v in missing_values:
            d[t] = null
        elif map_column2type[t] in (int, float):
            d[t] = str(row[t])
        else:
            d[t] = string_value % row[t]

    return d


def quoteTableName(name, quote_char="_", backend="sqlite"):
    if backend == "sqlite":
        # no special characters. Column names can not start with a number.
        if name[0] in "0123456789":
            name = "_" + name
        return re.sub("[-(),\[\].:]", "_", name)
    elif backend in ("mysql", "postgres"):
        if name[0] in "0123456789":
            name = "_" + name
        return re.sub("[-(),\[\]]:", "_", name)


def createTable(dbhandle,
                error,
                tablename,
                options,
                retry=True,
                ignore_empty=True,
                ignore_columns=[],
                rename_columns=[],
                lowercase=False,
                ignore_duplicates=True,
                indices=[],
                rows=None,
                headers=None,
                first_column=None,
                existing_tables=set(),
                append=False):

    # create table by guessing column types from data type.
    if rows:
        map_column2type, ignored, max_values = CSV.getMapColumn2Type(
            rows,
            ignore_empty=ignore_empty,
            get_max_values=True)
        if ignored:
            E.info("ignored columns: %s" % str(ignored))

        headers = list(map_column2type.keys())
        headers.sort()

    elif headers:
        map_column2type = dict(list(zip(headers, [None, ] * len(headers))))
        ignored = 0

    columns_to_ignore = set([x.lower() for x in ignore_columns])
    columns_to_rename = dict([x.lower().split(":")
                              for x in rename_columns])

    take = []
    # associate headers to field names
    columns = []
    present = {}
    for header_index, h in enumerate(headers):
        hh = h
        if lowercase:
            hh = string.lower(h)

        if hh in columns_to_ignore:
            continue

        if hh in present:
            if ignore_duplicates:
                continue
            else:
                raise ValueError("duplicate column %s" % hh)

        present[hh] = 1
        take.append(h)
        if map_column2type[h] == int:
            max_value = max_values[h]
            if max_value > 2147483647:
                t = "BIGINT DEFAULT '0'"
            elif max_value > 32767:
                t = "INTEGER DEFAULT '0'"
            else:
                t = "SMALLINT DEFAULT '0'"

        elif map_column2type[h] == float:
            t = "FLOAT DEFAULT '0'"
        else:
            if h in options.indices:
                t = options.index
            else:
                t = options.text

        # remove special characters from column names
        if hh == "":
            if first_column is not None and header_index == 0:
                hh = first_column
            else:
                raise ValueError("column '%s' without header " % h)
        hh = columns_to_rename.get(hh, hh)
        hh = re.sub('''['"]''', "", hh)
        hh = re.sub("[,;.:\-\+/ ()%?]", "_", hh)
        if hh[0] in "0123456789":
            hh = "_" + hh
        columns.append("%s %s" % (hh, t))

    if not options.append:
        # delete old table if it exists
        while 1:
            try:
                cc = dbhandle.cursor()
                # mysql: removed '' around table name
                statement = "DROP TABLE IF EXISTS %s" % tablename
                E.debug(statement)
                cc.execute(statement)
                dbhandle.commit()
                cc.close()
                E.info("existing table %s deleted" % tablename)
            except sqlite3.OperationalError as msg:
                E.warn(msg)
                time.sleep(5)
                continue
            except error as msg:
                E.warn("could not delete existing table %s: %s" %
                       (tablename, str(msg)))
                dbhandle.rollback()
                if not retry:
                    raise error(msg)
                elif tablename in existing_tables:
                    # table exists, but drop did not work (e.g. database lock)
                    time.sleep(5)
                    continue
                else:
                    # table might not have existed
                    break
            break

        # create new table
        statement = "CREATE TABLE %s ( %s );" % (
            tablename, ", ".join(columns))

        E.debug("table create:\n# %s" % (statement))

        while 1:
            try:
                cc = dbhandle.cursor()
                cc.execute(statement)
                cc.close()
                dbhandle.commit()
            except error as msg:
                E.warn("table creation failed: msg=%s, statement=\n  %s" %
                       (msg, statement))
                # TODO: check for database locked msg
                if not retry:
                    raise error(msg)
                if not re.search("locked", str(msg)):
                    raise error("%s: %s" % (msg, statement))
                time.sleep(5)
                continue
            break

        E.info("table %s created successfully." % tablename)

    return take, map_column2type, ignored


def run(infile, options, report_step=10000):

    options.tablename = quoteTableName(
        options.tablename, backend=options.backend)

    if options.map:
        m = {}
        for x in options.map:
            f, t = x.split(":")
            m[f] = t
        options.map = m
    else:
        options.map = {}

    existing_tables = set()

    quick_import_separator = "\t"

    if options.database_backend == "postgres":
        import psycopg2
        raise NotImplementedError("needs refactoring for commandline options")
        dbhandle = psycopg2.connect(options.psql_connection)
        error = psycopg2.Error
        options.null = "NULL"
        options.string_value = "'%s'"
        options.text = "TEXT"
        options.index = "TEXT"
        if options.insert_quick:
            raise ValueError("quick import not implemented.")

    elif options.database_backend == "mysql":
        import MySQLdb
        dbhandle = MySQLdb.connect(host=options.database_host,
                                   user=options.database_username,
                                   passwd=options.database_password,
                                   port=options.database_port,
                                   db=options.database_name)
        error = Exception
        options.null = "NULL"
        options.string_value = "%s"
        options.text = "TEXT"
        options.index = "VARCHAR(40)"
        if options.insert_quick:
            raise ValueError("quick import not implemented.")

    elif options.backend == "sqlite":
        import sqlite3
        dbhandle = sqlite3.connect(options.database_name)
        try:
            os.chmod(options.database_name, 0o664)
        except OSError as msg:
            E.warn("could not change permissions of database: %s" % msg)

        # Avoid the following error:
        # sqlite3.ProgrammingError: You must not use 8-bit bytestrings
        # unless you use a text_factory that can interpret 8-bit
        # bytestrings (like text_factory = str). It is highly
        # recommended that you instead just switch your application
        # to Unicode strings
        # Note: might be better to make csv2db unicode aware.
        dbhandle.text_factory = str

        error = sqlite3.OperationalError
        options.insert_many = True  # False
        options.null = None  # "NULL"
        options.text = "TEXT"
        options.index = "TEXT"
        options.string_value = "%s"  # "'%s'"

        statement = "SELECT name FROM sqlite_master WHERE type='table'"
        cc = executewait(dbhandle, statement, error, options.retry)
        existing_tables = set([x[0] for x in cc])
        cc.close()

        # use , as separator
        quick_import_statement = \
            "sqlite3 %s '.import %%s %s'" % \
            (options.database_name, options.tablename)

        quick_import_separator = "|"

    if options.header is not None:
        options.header = [x.strip() for x in options.header.split(",")]

    if options.utf:
        reader = CSV.UnicodeDictReader(infile,
                                       dialect=options.dialect,
                                       fieldnames=options.header)
    else:
        reader = csv.DictReader(CSV.CommentStripper(infile),
                                dialect=options.dialect,
                                fieldnames=options.header)

    if options.replace_header:
        try:
            next(reader)
        except StopIteration:
            pass

    E.info("reading %i columns to guess column types" % options.guess_size)

    rows = []
    for row in reader:
        if None in row:
            raise ValueError(
                "undefined columns in input file at row: %s" % row)

        try:
            rows.append(IOTools.convertDictionary(row, map=options.map))
        except TypeError as msg:
            E.warn(
                "incomplete line? Type error in conversion: "
                "'%s' with data: %s" % (msg, str(row)))
        except ValueError as msg:
            E.warn(
                "incomplete line? Type error in conversion: "
                "'%s' with data: %s" % (msg, str(row)))

        if len(rows) >= options.guess_size:
            break

    E.info("read %i rows for type guessing" % len(rows))
    E.info("creating table")

    if len(rows) == 0:
        if options.allow_empty:
            if not reader.fieldnames:
                E.warn("no data - no table created")
            else:
                # create empty table and exit
                take, map_column2type, ignored = createTable(
                    dbhandle,
                    error,
                    options.tablename,
                    options,
                    retry=options.retry,
                    headers=reader.fieldnames,
                    ignore_empty=options.ignore_empty,
                    ignore_columns=options.ignore_columns,
                    rename_columns=options.rename_columns,
                    lowercase=options.lowercase,
                    ignore_duplicates=options.ignore_duplicates,
                    indices=options.indices,
                    first_column=options.first_column,
                    existing_tables=existing_tables,
                    append=options.append)
                E.info("empty table created")
            return
        else:
            raise ValueError("empty table")
    else:
        take, map_column2type, ignored = createTable(
            dbhandle,
            error,
            options.tablename,
            options,
            rows=rows,
            retry=options.retry,
            headers=reader.fieldnames,
            ignore_empty=options.ignore_empty,
            ignore_columns=options.ignore_columns,
            rename_columns=options.rename_columns,
            lowercase=options.lowercase,
            ignore_duplicates=options.ignore_duplicates,
            indices=options.indices,
            first_column=options.first_column,
            existing_tables=existing_tables,
            append=options.append)

    def row_iter(rows, reader):
        for row in rows:
            yield quoteRow(row, take, map_column2type,
                           options.missing_values,
                           null=options.null,
                           string_value=options.string_value)
        for data in reader:
            yield quoteRow(IOTools.convertDictionary(data, map=options.map),
                           take,
                           map_column2type,
                           options.missing_values,
                           null=options.null,
                           string_value=options.string_value)

    ninput = 0

    E.info("inserting data")

    if options.insert_quick:
        E.info("using quick insert")

        outfile, filename = tempfile.mkstemp()

        E.debug("dumping data into %s" % filename)

        for d in row_iter(rows, reader):

            ninput += 1
            os.write(outfile, quick_import_separator.join(
                [str(d[x]) for x in take]) + "\n")

            if ninput % report_step == 0:
                E.info("iteration %i\n" % ninput)

        os.close(outfile)

        statement = quick_import_statement % filename
        E.debug(statement)

        # infinite loop possible
        while 1:

            retcode = E.run(statement, cwd=os.getcwd(), close_fds=True)

            if retcode != 0:
                E.warn("import error using statement: %s" % statement)

                if not options.retry:
                    raise ValueError(
                        "import error using statement: %s" % statement)

                time.sleep(5)
                continue

            break

        os.remove(filename)

        # there is no way to insert NULL values into sqlite. The only
        # solution is to update all colums.
        for column in take:
            executewait(dbhandle,
                        "UPDATE %s SET %s = NULL WHERE %s = 'None'" % (
                            options.tablename, column, column),
                        error,
                        options.retry)

    elif options.insert_many:
        data = []
        for d in row_iter(rows, reader):
            ninput += 1

            data.append([d[x] for x in take])

            if ninput % report_step == 0:
                E.info("iteration %i" % ninput)

        statement = "INSERT INTO %s VALUES (%s)" % (
            options.tablename, ",".join("?" * len(take)))

        E.info("inserting %i rows" % len(data))
        E.debug("multiple insert:\n# %s" % statement)

        while 1:
            try:
                dbhandle.executemany(statement, data)
            except error as msg:
                E.warn("import failed: msg=%s, statement=\n  %s" %
                       (msg, statement))
                # TODO: check for database locked msg
                if not options.retry:
                    raise error(msg)
                if not re.search("locked", str(msg)):
                    raise error(msg)
                time.sleep(5)
                continue
            break

    else:
        # insert line by line (could not figure out how to do bulk loading with
        # subprocess and COPY FROM STDIN)
        statement = "INSERT INTO %s VALUES (%%(%s)s)" % (options.tablename,
                                                         ')s, %('.join(take))
        # output data used for guessing:
        for d in row_iter(rows, reader):

            ninput += 1
            E.debug("single insert:\n# %s" % (statement % d))
            cc = executewait(dbhandle, statement, error,
                             retry=options.retry,
                             args=d)
            cc.close()

            if ninput % report_step == 0:
                E.info("iteration %i" % ninput)

    E.info("building indices")
    nindex = 0
    for index in options.indices:

        nindex += 1
        try:
            statement = "CREATE INDEX %s_index%i ON %s (%s)" % (
                options.tablename, nindex, options.tablename, index)
            cc = executewait(dbhandle, statement, error, options.retry)
            cc.close()
            E.info("added index on column %s" % (index))
        except error as msg:
            E.info("adding index on column %s failed: %s" % (index, msg))

    statement = "SELECT COUNT(*) FROM %s" % (options.tablename)
    cc = executewait(dbhandle, statement, error, options.retry)
    result = cc.fetchone()
    cc.close()

    noutput = result[0]

    E.info("ninput=%i, noutput=%i, nskipped_columns=%i" %
           (ninput, noutput, len(ignored)))

    dbhandle.commit()


def buildParser():

    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option("--csv-dialect", dest="dialect", type="string",
                      help="csv dialect to use [default=%default].")

    parser.add_option(
        "-m", "--map", dest="map", type="string", action="append",
        help="explicit mapping function for columns The format is "
        "column:type (e.g.: length:int) [default=%default].")

    parser.add_option("-t", "--table", dest="tablename", type="string",
                      help="table name for all backends [default=%default].")

    parser.add_option("-d", "--database", dest="database", type="string",
                      help="database name for sqlite3 [default=%default].")

    parser.add_option(
        "-H", "--header-names", dest="header", type="string",
        help="',' separated list of column headers for files without "
        "column header [default=%default].")

    parser.add_option("--replace-header", dest="replace_header",
                      action="store_true",
                      help="replace header with --header-names instead of "
                      "adding it [default=%default].")

    parser.add_option("-l", "--lowercase-fields", dest="lowercase",
                      action="store_true",
                      help="force lower case column names "
                      "[default=%default].")

    parser.add_option("-u", "--ignore-duplicates", dest="ignore_duplicates",
                      action="store_true",
                      help="ignore columns with duplicate names "
                      "[default=%default].")

    parser.add_option("-s", "--ignore-same", dest="ignore_same",
                      action="store_true",
                      help="ignore columns with identical values "
                      "[default=%default].")

    parser.add_option("--ignore-column", dest="ignore_columns", type="string",
                      action="append",
                      help="ignore columns [default=%default].")

    parser.add_option("--rename-column", dest="rename_columns", type="string",
                      action="append",
                      help="rename columns [default=%default].")

    parser.add_option("--first-column", dest="first_column", type="string",
                      help="name of first column - permits loading CSV "
                      "table where the first "
                      "column name is the empty string [default=%default].")

    parser.add_option("-e", "--ignore-empty", dest="ignore_empty",
                      action="store_true",
                      help="ignore columns which are all empty "
                      "[default=%default].")

    parser.add_option("-q", "--quick", dest="insert_quick",
                      action="store_true",
                      help="try quick file based import - needs to "
                      "be supported by the backend [default=%default].")

    parser.add_option("-i", "--add-index", dest="indices", type="string",
                      action="append",
                      help="create an index for the named column "
                      "[default=%default].")

    parser.add_option("-a", "--allow-empty-file", dest="allow_empty",
                      action="store_true",
                      help="allow empty table [default=%default].")

    parser.add_option("--retry", dest="retry", action="store_true",
                      help="retry if an SQL statement fails - warning: "
                      "THIS MIGHT CAUSE DEADLOCKS [default=%default].")

    parser.add_option("-z", "--from-zipped", dest="from_zipped",
                      action="store_true",
                      help="input is zipped.")

    parser.add_option("--append", dest="append",
                      action="store_true",
                      help="append to existing table [default=%default].")

    parser.add_option(
        "--utf8", dest="utf", action="store_true",
        help="standard in is encoded as UTF8 rather than local default"
        ", WARNING: does not strip comment lines yet [default=%default]")

    parser.set_defaults(
        map=[],
        dialect="excel-tab",
        database="csvdb",
        lowercase=False,
        tablename="csv",
        from_zipped=False,
        ignore_duplicates=False,
        ignore_identical=False,
        ignore_empty=False,
        insert_many=False,
        ignore_columns=[],
        rename_columns=[],
        header=None,
        replace_header=False,
        guess_size=1000,
        report_step=10000,
        backend="sqlite",
        indices=[],
        missing_values=("na", "NA", ),
        insert_quick=False,
        allow_empty=False,
        retry=False,
        utf=False,
        append=False,
    )

    return parser



'''fastq2table.py - compute stats on reads in fastq files
======================================================

:Tags: Genomics NGS Sequences FASTQ Annotation

Purpose
-------

This script iterates over a fastq file and outputs
summary statistics for each read.

The output is a tab-delimited text file with the following columns:

+----------+----------------------------------------+
|*Column*  |*Content*                               |
+----------+----------------------------------------+
|read      |read identifier present in input fastq  |
|          |file                                    |
+----------+----------------------------------------+
|nfailed   |number of reads that fall below Q10     |
+----------+----------------------------------------+
|nN        |number of ambiguous base calls (N)      |
+----------+----------------------------------------+
|nval      |number of bases in the read             |
+----------+----------------------------------------+
|min       |minimum base quality score for the read |
+----------+----------------------------------------+
|max       |maximum base quality for the read       |
+----------+----------------------------------------+
|mean      |mean base quality for the read          |
+----------+----------------------------------------+
|median    |median base quality for the read        |
+----------+----------------------------------------+
|stddev    |standard devitation of quality scores   |
|          |for the read                            |
+----------+----------------------------------------+
|sum       |sum of quality scores for the read      |
+----------+----------------------------------------+
|q1        |25th percentile of quality scores for   |
|          |the read                                |
+----------+----------------------------------------+
|q3        |25th percentile of quality scores for   |
|          |the read                                |
+----------+----------------------------------------+

Usage
-----

Example::

   cgat fastq2table --guess-format=sanger < in.fastq > out.tsv

In this example we know that our data have quality scores formatted as
sanger. Given that illumina-1.8 quality scores are highly overlapping
with sanger, this option defaults to sanger qualities. In default mode
the script may not be able to distinguish highly overlapping sets of
quality scores.

If we provide two reads to the script::

   @DHKW5DQ1:308:D28FGACXX:5:2211:8051:4398
   ACAATGTCCTGATGTGAATGCCCCTACTATTCAGATCGCTTAGGGCATGC
   +
   B1=?DFDDHHFFHIJJIJGGIJGFIEE9CHIIFEGGIIJGIGIGIIDGHI
   @DHKW5DQ1:308:D28FGACXX:5:1315:15039:83265
   GAATGCCCCTACTATTCAGATCGCTTAGGGCATGCGTCGCATGTGAGTAA
   +
   @@@FDFFFHGHHHJIIIJIGHIJJIGHGHC9FBFBGHIIEGHIGC>F@FA

we get the following table as output:

+-----------------------------------------+-------+--+----+-------+-------+-------+-------+------+---------+-------+-------+
|read                                     |nfailed|nN|nval|min    |max    |mean   |median |stddev|sum      |q1     |q3     |
+-----------------------------------------+-------+--+----+-------+-------+-------+-------+------+---------+-------+-------+
|DHKW5DQ1:308:D28FGACXX:5:2211:8051:4398  |0      |0 |50  |16.0000|41.0000|37.2000|38.0000|4.4900|1860.0000|36.0000|40.0000|
+-----------------------------------------+-------+--+----+-------+-------+-------+-------+------+---------+-------+-------+
|DHKW5DQ1:308:D28FGACXX:5:1315:15039:83265|0      |0 |50  |24.0000|41.0000|37.0200|38.0000|3.5916|1851.0000|36.0000|40.0000|
+-----------------------------------------+-------+--+----+-------+-------+-------+-------+------+---------+-------+-------+


Type::

   cgat fastq2table --help

for command line help.

Command line options
--------------------

'''

import sys

import CGAT.Experiment as E
import CGAT.Stats as Stats
import CGAT.Fastq as Fastq


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option(
        "--guess-format", dest="guess_format", type="choice",
        choices=(
            'sanger', 'solexa', 'phred64', 'illumina-1.8', 'integer'),
        help="The default behaviour of the script is to guess the quality "
        "format of the input fastq file. The user can specify the "
        "quality format of the input file using the --guess-format option. "
        "The script will use this format if the "
        "sequence qualities are ambiguous.[default=%default].")

    parser.add_option(
        "--target-format", dest="target_format", type="choice",
        choices=(
            'sanger', 'solexa', 'phred64', 'illumina-1.8', 'integer'),
        help="The script will convert quality scores to the destination "
        "format unless [default=%default].")

    parser.set_defaults(
        target_format=None,
        guess_format=None,
        min_quality=10,
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)

    c = E.Counter()

    if options.target_format:
        iterator = Fastq.iterate_convert(options.stdin,
                                         format=options.target_format,
                                         guess=options.guess_format)
    else:
        iterator = Fastq.iterate_guess(options.stdin,
                                       guess=options.guess_format)

    options.stdout.write("read\tnfailed\tnN\t%s\n" %
                         ("\t".join(Stats.Summary().getHeaders())))

    min_quality = options.min_quality

    for record in iterator:
        c.input += 1
        quals = record.toPhred()
        nfailed = len([x for x in quals if x < min_quality])
        nns = record.seq.count("N") + record.seq.count(".")
        options.stdout.write("%s\t%i\t%i\t%s\n" % (record.identifier,
                                                   nfailed,
                                                   nns,
                                                   str(Stats.Summary(quals))
                                                   ))
        c.output += 1

    # write footer and output benchmark information.
    E.info("%s" % str(c))
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''diff_bam.py - compare multiple bam files against each other
===========================================================

:Tags: Genomics NGS BAM Comparison

Purpose
-------

Compare reads in multiple BAM files against each other.

.. note::

    BAM files need to be sorted by read name. samtools sort does NOT
    work as it uses a custom comparison function (strnum_cmp) that is
    incompatible with the standard lexicographical order in
    python. See the example below on how to get sorted files.

This script is for validation purposes. It might take a while
for large BAM files.

Usage
-----

If you have two sorted :term:`sam` or :term:`bam` formatted
files, type::

   cgat diff_bam a.bam b.bam > out

If they are not sorted, you can use samtools sort to do an
inplace sort::

   cgat diff_bam <(samtools view -h a.bam | hsort 0 -k1,1)
                  <(samtools view -h b.bam | hsort 0 -k1,1)

The samtools ``-h`` option outputs the header, and the hsort command
sorts without disturbing the header.

An example output looks like this:

+------------------------------------+----------+--------+--------+--------+---------+---------+
|read                                |nlocations|nmatched|file1_nh|file2_nh|file1_loc|file2_loc|
+------------------------------------+----------+--------+--------+--------+---------+---------+
|42YKVAAXX_HWI-EAS229_1:1:11:1659:174|1         |2       |2       |2       |0,0      |0,0      |
+------------------------------------+----------+--------+--------+--------+---------+---------+
|42YKVAAXX_HWI-EAS229_1:1:11:166:1768|1         |2       |1       |1       |0        |0        |
+------------------------------------+----------+--------+--------+--------+---------+---------+
|612UOAAXX_HWI-EAS229_1:1:97:147:1248|2         |2       |2       |2       |0,1      |0,1      |
+------------------------------------+----------+--------+--------+--------+---------+---------+

This reports for each read the number of locations that the read maps to
in all files, the number of files that have matches found for the read.
Then, for each file, it reports the number of matches and the locations
it maps to (coded as integers, 0 the first location, 1 the second, ...).

In the example above, the first read maps twice to 1 location in both
files.  This is a read occuring twice in the input file. The second
read maps to the same one location in both files, while the third read
maps to the two same locations in both input files.

Type::

   python diff_bam.py --help

for command line help.

Documentation
-------------

For read counts to be correct the NH flag to be set correctly.

Command line options
--------------------

'''

import sys
import itertools
import pysam
import CGAT.Experiment as E


class multiway_groupby(object):
    # [k for k, g in groupby('AAAABBBCCDAABBB')] --> A B C D A B
    # [list(g) for k, g in groupby('AAAABBBCCD')] --> AAAA BBB CC D

    def __init__(self, iterables, key=None):
        # set up iterators
        self.it = [itertools.groupby(iterable, key) for iterable in iterables]

        # todo: catch StopIterator for empty iterables
        self.current = [next(self.it[x]) for x in range(len(iterables))]

        # decide on target key
        self.targetkey = min([x[0] for x in self.current if x[0] is not None])

    def __iter__(self):
        return self

    def __next__(self):

        # check if all iterators exhausted
        if not any(self.it):
            raise StopIteration

        # yield all that correspond to target key and update
        result = []
        targetkey = self.targetkey
        updated = 0

        for x, y in enumerate(self.current):
            key, val = y
            if key and key == targetkey:
                # save result (instantiate to prevent overwriting)
                # in next method
                result.append(list(val))
                # advance
                try:
                    self.current[x] = next(self.it[x])
                except StopIteration:
                    self.it[x] = None
                    self.current[x] = (None, None)
                updated += 1
            else:
                # return empty result
                result.append([])

        assert updated > 0, "no updates - infinite loop"

        # decide which is target key
        try:
            self.targetkey = min([x[0]
                                  for x in self.current if x[0] is not None])
        except ValueError:
            # if all are None, sequence is empty
            self.targetkey = None

        return targetkey, result

    def next(self):
        return self.__next__()


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option(
        "--header-names", dest="headers", type="string",
        help="',' separated list of labels used as headers. "
        " Should correspond in order to command line arguments [%default]")

    parser.set_defaults(
        headers=None,
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv, add_output_options=True)

    read_length = 50
    max_distance = read_length

    if len(args) < 2:
        raise ValueError("please specify at least two BAM files")

    infiles = []
    for arg in args:
        infiles.append(pysam.AlignmentFile(arg, 'rb'))

    if options.headers:
        headers = options.headers.split(",")
        if len(headers) != len(args):
            raise ValueError("number of headers and files differrent")
    else:
        headers = ["file%i" % x for x in range(1, len(infiles) + 1)]

    options.stdout.write("read\tnlocations\tnmatched\t%s\t%s\n" %
                         ("\t".join(["%s_nh" % x for x in headers]),
                          "\t".join(["%s_loc" % x for x in headers])))

    ninput, noutput = 0, 0
    for readname, result in multiway_groupby(infiles, key=lambda x: x.qname):
        ninput += 1
        nmatches = 0
        locations = set()

        # build location directory
        for r in result:
            for rr in r:
                if rr.is_unmapped:
                    continue
                locations.add((rr.tid, rr.pos))

        # permit a certain fuzzyness in locations
        # as some mappers clip and others don't.
        locations = list(locations)
        locations.sort()
        pos2loc = {}
        last_tid, last_pos, nlocations = None, 0, -1
        for tid, pos in locations:
            if tid != last_tid or pos - last_pos > max_distance:
                nlocations += 1
            pos2loc[(tid, pos)] = nlocations
            last_tid, last_pos = tid, pos
        nlocations += 1

        # code locations
        codes, nh = [], []
        for r in result:
            c = []
            for rr in r:
                if rr.is_unmapped:
                    continue
                c.append(pos2loc[(rr.tid, rr.pos)])
            codes.append(",".join(map(str, sorted(c))))
            nh.append(len(c))
            if len(c):
                nmatches += 1

        noutput += 1
        options.stdout.write("%s\t%i\t%i\t%s\t%s\n" % (
            readname,
            nlocations,
            nmatches,
            "\t".join(
                ["%i" % x for x in nh]),
            "\t".join(codes)))

    E.info("ninput=%i, noutput=%i" % (ninput, noutput))

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""FastaIterator.py - Iteration over fasta files
================================================

This module provides a simple iterator of Fasta formatted files.  The
difference to the biopython iterator is that the iterators in this
module skip over comment lines starting with "#".

.. note::
   Another way to access the information in :term:`fasta` formatted
   files is through pysam_.

Reference
---------

"""
import subprocess
import os


class FastaRecord:
    """a :term:`fasta` record.

    Attributes
    ----------
    title: string
       the title of the sequence

    sequence: string
       the sequence

    fold : int
       the number of bases per line when writing out
    """

    def __init__(self, title, sequence, fold=False):

        self.title = title
        self.sequence = sequence
        self.fold = fold

    def __str__(self):
        ''' str method for writing out'''

        if self.fold:
            seq = [self.sequence[i:i + self.fold]
                   for i in range(0, len(self.sequence), self.fold)]
        else:
            seq = (self.sequence,)

        return ">%s\n%s" % (self.title, "\n".join(seq))


class FastaIterator:
    '''a iterator of :term:`fasta` formatted files.

    Yields
    ------
    FastaRecord

    '''

    def __init__(self, f, *args, **kwargs):
        self.iterator = iterate(f)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterator)

    def next(self):
        return next(self.iterator)


def iterate(infile, comment="#", fold=False):
    '''iterate over fasta data in infile

    Lines before the first fasta record are
    ignored (starting with ``>``) as well as
    lines starting with the comment character.

    Parameters
    ----------
    infile : File
        the input file
    comment : char
        comment character
    fold : int
        the number of bases before line split when writing out

    Yields
    ------
    FastaRecord
    '''
    h = infile.readline()[:-1]

    if not h:
        raise StopIteration

    # skip everything until first fasta entry starts
    while h[0] != ">":
        h = infile.readline()[:-1]
        if not h:
            raise StopIteration
        continue

    h = h[1:]
    seq = []

    for line in infile:

        if line.startswith(comment):
            continue

        if line.startswith('>'):
            yield FastaRecord(h, ''.join(seq), fold)

            h = line[1:-1]
            seq = []
            continue

        seq.append(line[:-1])

    yield FastaRecord(h, ''.join(seq), fold)


def iterate_together(*args):
    """iterate synchronously over one or more fasta files.

    The iteration finishes once any of the files is exhausted.

    Arguments
    ---------

    :term:`fasta`-formatted files to be iterated upon

    Yields
    ------
    tuple
       a tuple of :class:`FastaRecord` corresponding to
       the current record in each file.
    """

    iterators = [FastaIterator(x) for x in args]

    while 1:
        yield [next(x) for x in iterators]


def count(filename):
    '''count number of sequences in fasta file.

    This method uses the ``grep`` utility to count
    lines starting with ``>``.

    Arguments
    ---------
    filename : string
        The filename

    Raises
    ------
    OSError
        If the file does not exist

    Returns
    -------
    int
        The number of sequences in the file.
    '''

    if filename.endswith(".gz"):
        statement = "zcat %s | grep -c '>'" % filename
    else:
        statement = "cat %s | grep -c '>'" % filename

    if not os.path.exists(filename):
        raise OSError("file '%s' does not exist" % filename)

    # grep returns error if no match is found
    try:
        return subprocess.check_output(statement, shell=True)
    except subprocess.CalledProcessError:
        return 0



'''bam2geneprofile.py - build meta-gene profile for a set of transcripts/genes
===========================================================================

:Tags: Genomics NGS Genesets Intervals GTF BAM Summary

Purpose
-------

This script takes a :term:`gtf` formatted file, a short reads
:term:`bam` formatted file and computes meta-gene profiles over
various annotations derived from the :term:`gtf` file.

A meta-gene profile is an abstract genomic entity over which reads
stored in a :term:`bam` formatted file have been counted. A meta-gene
might be an idealized eukaryotic gene (upstream, exonic sequence,
downstream) or any other genomic landmark of interest such as
transcription start sites.

The script can be used to visualize binding profiles of a chromatin
mark in gene bodies, binding of transcription factors in promotors or
sequencing bias (e.g. 3' bias) in RNA-Seq data.

This script is designed with a slight emphasis on RNA-Seq
datasets. For example, it takes care of spliced reads, by using the
CIGAR string in the BAM file to accurately define aligned bases (if
the --base-accurate is specified, currently this feature is turned off
by default for speed reasons).

Alternatively, for the purpose of visualizing binding profiles of
transcription factor ChIP-Seq without the need to use any genomic
annotations (ENSEMBL, or refseq), you may also consider using
:doc:`bam2peakshape`, which is designed with a slight emphasis on
Chip-Seq datasets. For example, :doc:`bam2peakshape` is able to center
the counting window to the summit of every individual peak.
:doc:`bam2peakshape` is also able to: (1) plot the control ChIP-Seq
library to enable side-by-side comparison; (2) randomize the given
regions to provide a semi-control.

Usage
-----

Quick start examples
++++++++++++++++++++

The following command will generate the gene profile plot similar to
Fig 1(a) in the published CGAT paper, but using a test dataset that is
much smaller and simpler than the dataset used for publishing the CGAT
paper. ::

    python ./scripts/bam2geneprofile.py
        --bam-file=./tests/bam2geneprofile.py/multipleReadsSplicedOutAllIntronsAndSecondExon.bam
        --gtf-file=./tests/bam2geneprofile.py/onegeneWithoutAnyCDS.gtf.gz
        --method=geneprofile
        --reporter=gene

In the following, a slightly more involved example will use more
features of this script. The following command generate the gene
profile showing base accuracy of upstream (500bp), exons, introns and
downstream(500bp) of a gene model from some user supplied RNA-Seq data
and geneset. ::

    python ./scripts/bam2geneprofile.py
        --bam-file=./rnaseq.bam
        --gtf-file=./geneset.gtf.gz
        --method=geneprofilewithintrons
        --reporter=gene
        --extension-upstream=500
        --resolution-upstream=500
        --extension-downstream=500
        --resolution-downstream=500

The output will contain read coverage over genes. The profile will
contain four separate segments:

1. the upstream region of a gene ( set to be 500bp ),
   (``--extension-upstream=500``).

2. the transcribed region of a gene. The transcribed region of every gene will
   be scaled to 1000 bp (default), shrinking longer transcripts and
   expanding shorter transcripts.

3. the intronic regions of a gene. These will be scaled to 1000b (default).

4. the downstream region of a gene (set to be 500bp),
   (``--extension-downstream=500``).



Detailed explaination
+++++++++++++++++++++

The :file:`bam2geneprofile.py` script reads in a set of transcripts
from a :term:`gtf` formatted file. For each transcript, overlapping
reads from the provided :term:`bam` file are collected. The counts
within the transcript are then mapped onto the meta-gene structure and
counts are aggregated over all transcripts in the :term:`gtf` file.

:term:`Bam` files need to be sorted by coordinate and indexed.

A meta-gene structure has two components - regions of variable size,
such as exons, introns, etc, which nevertheless have a fixed start and
end coordinate in a transcript. The other component are regions of
fixed width, such a regions of a certain size upstream or downstream
of a landmark such as a transcription start site.

The size of the former class, regions of variable size, can be varied
with ``--resolution`` options. For example, the option
``--resolution-upstream-utr=1000`` will create a meta-gene with a
1000bp upstream UTR region. UTRs that are larger will be compressed,
and UTRs that are smaller, will be stretched to fit the 1000bp
meta-gene UTR region.

The size of fixed-width regions can be set with ``--extension``
options. For example, the options ``--extension-upstream`` will set
the size of the uptsream extension region to 1000bp. Note that no
scaling is required when counting reads towards the fixed-width
meta-gene profile.

Type::

   python bam2geneprofile.py --help

for command line help.

Options
-------

The script provides a variety of different meta-gene structures i.e.
geneprofiles, selectable via using the option: (``--method``).

Profiles
++++++++

Different profiles are accessible through the ``--method`` option. Multiple
methods can be applied at the same time. While ``upstream`` and ``downstream``
typically have a fixed size, the other regions such as ``CDS``, ``UTR`` will be
scaled to a common size.

utrprofile
    UPSTREAM - UTR5 - CDS - UTR3 - DOWNSTREAM
    gene models with UTR. Separate the coding section from the non-coding part.

geneprofile
    UPSTREAM - EXON - DOWNSTREAM
    simple exonic gene models

geneprofilewithintrons
    UPSTREAM - EXON - INTRON - DOWNSTREAM

    gene models containing also intronic sequence, only correct if
    used with ``--use-base-accuracy`` option.

separateexonprofile
    UPSTREAM - FIRST EXON - EXON - LAST EXON - DOWNSTREAM

    gene models with the first and last exons separated out from all
    other exons.  Only applicable to gene models with > 1 exons.

separateexonprofilewithintrons
    UPSTREAM - FIRST EXON - EXON - INTRON - LAST EXON - DOWNSTREAM

    gene models with first and last exons separated out, and includes
    all introns together.  Excludes genes with < 2 exons and no introns.

geneprofileabsolutedistancefromthreeprimeend

    UPSTREAM - EXON (absolute distance, see below) - INTRON (absolute
    distance, see below) - DOWNSTREAM (the downstream of the exons)
    region, the script counts over the mRNA transcript only, skipping
    introns. Designed to visualize the 3 prime bias in RNASeq data,
    only correct if used together with ``--use-base-accuracy`` option.

    absolute distance: In order to to visualize the 3 prime bias,
    genes are not supposed to be streched to equal length as it did in
    all other counting methods. In this counting method, we first set
    a fix length using
    ``--extension-exons-absolute-distance-topolya``, the script will
    discard genes shorter than this fixed length. For genes (when all
    the exons stitched together) longer than this fixed length, the
    script will only count over this fixed length ( a absolute
    distance ) from three prime end, instead of compress the longer
    genes. Same goes for absolute distance intron counting.

tssprofile
    UPSTREAM - DOWNSTREAM
    transcription start/stop sites

intervalprofile
    UPSTREAM - INTERVAL - DOWNSTREAM
    Similar to geneprofile, but count over the complete span of the gene
    (including introns).

midpointprofile
    UPSTREAM  - DOWNSTREAM
    aggregate over midpoint of gene model


Normalization
+++++++++++++

Normalization can be applied in two stages of the computation.

Count vector normalization
^^^^^^^^^^^^^^^^^^^^^^^^^^

Before adding counts to the meta-gene profile, the profile for the
individual transcript can be normalized. Without normalization, highly
expressed genes will contribute more to the meta-gene profile than
lowly expressed genes.  Normalization can assure that each gene
contributes an equal amount.

Normalization is applied to the vector of read counts that is computed
for each transcript. Normalization can be applied for the whole
transcript (``total``) or on a per segment basis depending on the
counter. For example, in the gene counter, exons, upstream and
downstream segments can be normalized independently.

Counts can be normalized either by the maximum or the sum of all
counts in a segment or across the whole transcript. Normalization is
controlled with the command line option ``--normalize-trancript``. Its
arguments are:

* ``none``: no normalization
* ``sum``: sum of counts within a region (exons, upstream, ...).
  The area under the curve will sum to 1 for each region.
* ``max``: maximum count within a region (exons,upstream, ...).
* ``total-sum``: sum of counts across all regions. The area
  under the curve will sum to 1 for
  the complete transcript.
* ``total-max``: maximum count across all regions.

The options above control the contribution of individual transcripts
to a meta-gene profile and are thus suited for example for RNA-Seq data.

The options above do not control for different read-depths or any
local biases. To compare meta-gene profiles between samples,
additional normalization is required.

Meta-gene profile normalization
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

To enable comparison between experiments, the meta-gene profile itself
can be normalized.  Normalization a profile can help comparing the
shapes of profiles between different experiments independent of the
number of reads or transcripts used in the construction of the
meta-gene profile.

Meta-gene profile normalization is controlled via the
``--normalize-profile`` option. Possible normalization are:

* none: no normalization
* area: normalize such that the area under the meta-gene profile is 1.
* counts: normalize by number of features (genes,tss) that have been counted.
* background: normalize with background (see below).

A special normalization is activated with the ``background`` option.
Here, the counts at the left and right most regions are used to
estimate a background level for each transcript. The counts are then
divided by this background-level. The assumption is that the meta-gene
model is computed over a large enough area to include genomic
background.

Genes versus transcripts
++++++++++++++++++++++++

The default is to collect reads on a per-transcript
level. Alternatively, the script can merge all transcripts of a gene
into a single virtual transcript. Note that this virtual transcript
might not be a biologically plausible transcript. It is usually better
to provide :file:`bam2geneprofile.py` with a set of representative
transcripts per gene in order to avoid up-weighting genes with
multiple transcripts.

Control
+++++++

If control files (chip-seq input tracks) are supplied, counts in the
control file can be used to compute a fold-change.

Bed and wiggle files
++++++++++++++++++++

The densities can be computed from :term:`bed` or :term:`wiggle`
formatted files. If a :term:`bed` formatted file is supplied, it must
be compressed with and indexed with :file:`tabix`.

.. note::

   Paired-endedness is ignored. Both ends of a paired-ended read are
   treated individually.


Command line options
--------------------

'''

import os
import sys
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import pysam
import CGAT.GTF as GTF
import numpy
import pandas

import CGAT.scripts._bam2geneprofile as _bam2geneprofile


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
        """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("-m", "--method", dest="methods", type="choice",
                      action="append",
                      choices=("geneprofile", "tssprofile", "utrprofile",
                               "intervalprofile", "midpointprofile",
                               "geneprofilewithintrons",
                               "geneprofileabsolutedistancefromthreeprimeend",
                               "separateexonprofile",
                               "separateexonprofilewithintrons",
                               ),
                      help='counters to use. Counters describe the '
                      'meta-gene structure to use. '
                      'Note using geneprofilewithintrons, or '
                      'geneprofileabsolutedistancefromthreeprimeend will '
                      'automatically turn on the --use-base-accuracy option'
                      '[%default].')

    parser.add_option("-b", "--bam-file", "--bedfile", "--bigwigfile",
                      dest="infiles",
                      metavar="BAM",
                      type="string", action="append",
                      help="BAM/bed/bigwig files to use. Do not mix "
                      "different types [%default]")

    parser.add_option("-c", "--control-bam-file", dest="controlfiles",
                      metavar="BAM",
                      type="string", action="append",
                      help="control/input to use. Should be of the same "
                      "type as the bam/bed/bigwig file"
                      " [%default]")

    parser.add_option("-g", "--gtf-file", dest="gtffile", type="string",
                      metavar="GTF",
                      help="GTF file to use. "
                      "[%default]")

    parser.add_option(
        "--normalize-transcript",
        dest="transcript_normalization",
        type="choice",
        choices=("none", "max", "sum", "total-max", "total-sum"),
        help="normalization to apply on each transcript "
        "profile before adding to meta-gene profile. "
        "[%default]")

    parser.add_option(
        "--normalize-profile",
        dest="profile_normalizations",
        type="choice", action="append",
        choices=("all", "none", "area", "counts", "background"),
        help="normalization to apply on meta-gene "
        "profile normalization. "
        "[%default]")

    parser.add_option(
        "-r", "--reporter", dest="reporter", type="choice",
        choices=("gene", "transcript"),
        help="report results for genes or transcripts."
        " When 'genes` is chosen, exons across all transcripts for"
        " a gene are merged. When 'transcript' is chosen, counts are"
        " computed for each transcript separately with each transcript"
        " contributing equally to the meta-gene profile."
        " [%default]")

    parser.add_option("-i", "--shift-size", dest="shifts", type="int",
                      action="append",
                      help="shift reads in :term:`bam` formatted file "
                      "before computing densities (ChIP-Seq). "
                      "[%default]")

    parser.add_option("-a", "--merge-pairs", dest="merge_pairs",
                      action="store_true",
                      help="merge pairs in :term:`bam` formatted "
                      "file before computing "
                      "densities (ChIP-Seq). "
                      "[%default]")

    parser.add_option("-u", "--use-base-accuracy", dest="base_accuracy",
                      action="store_true",
                      help="compute densities with base accuracy. The default "
                      "is to only use the start and end of the aligned region "
                      "(RNA-Seq) "
                      "[%default]")

    parser.add_option("-e", "--extend", dest="extends", type="int",
                      action="append",
                      help="extend reads in :term:`bam` formatted file "
                      "(ChIP-Seq). "
                      "[%default]")

    parser.add_option("--resolution-upstream", dest="resolution_upstream",
                      type="int",
                      help="resolution of upstream region in bp "
                      "[%default]")

    parser.add_option("--resolution-downstream", dest="resolution_downstream",
                      type="int",
                      help="resolution of downstream region in bp "
                      "[%default]")

    parser.add_option("--resolution-upstream-utr",
                      dest="resolution_upstream_utr",
                      type="int",
                      help="resolution of upstream UTR region in bp "
                      "[%default]")

    parser.add_option("--resolution-downstream-utr",
                      dest="resolution_downstream_utr",
                      type="int",
                      help="resolution of downstream UTR region in bp "
                      "[%default]")

    parser.add_option("--resolution-cds", dest="resolution_cds", type="int",
                      help="resolution of cds region in bp "
                      "[%default]")

    parser.add_option("--resolution-first-exon", dest="resolution_first",
                      type="int",
                      help="resolution of first exon in gene, in bp"
                      "[%default]")

    parser.add_option("--resolution-last-exon", dest="resolution_last",
                      type="int",
                      help="resolution of last exon in gene, in bp"
                      "[%default]")

    parser.add_option("--resolution-introns",
                      dest="resolution_introns", type="int",
                      help="resolution of introns region in bp "
                      "[%default]")

    parser.add_option("--resolution-exons-absolute-distance-topolya",
                      dest="resolution_exons_absolute_distance_topolya",
                      type="int",
                      help="resolution of exons absolute distance "
                      "topolya in bp "
                      "[%default]")

    parser.add_option("--resolution-introns-absolute-distance-topolya",
                      dest="resolution_introns_absolute_distance_topolya",
                      type="int",
                      help="resolution of introns absolute distance "
                      "topolya in bp "
                      "[%default]")

    parser.add_option("--extension-exons-absolute-distance-topolya",
                      dest="extension_exons_absolute_distance_topolya",
                      type="int",
                      help="extension for exons from the absolute "
                      "distance from the topolya in bp "
                      "[%default]")

    parser.add_option(
        "--extension-introns-absolute-distance-topolya",
        dest="extension_introns_absolute_distance_topolya", type="int",
        help="extension for introns from the absolute distance from "
        "the topolya in bp [%default]")

    parser.add_option(
        "--extension-upstream", dest="extension_upstream", type="int",
        help="extension upstream from the first exon in bp"
        "[%default]")

    parser.add_option(
        "--extension-downstream", dest="extension_downstream", type="int",
        help="extension downstream from the last exon in bp"
        "[%default]")

    parser.add_option(
        "--extension-inward", dest="extension_inward", type="int",
        help="extension inward from a TSS start site in bp"
        "[%default]")

    parser.add_option(
        "--extension-outward", dest="extension_outward", type="int",
        help="extension outward from a TSS start site in bp"
        "[%default]")

    parser.add_option("--scale-flank-length", dest="scale_flanks", type="int",
                      help="scale flanks to (integer multiples of) gene length"
                      "[%default]")

    parser.add_option(
        "--control-factor", dest="control_factor", type="float",
        help="factor for normalizing control and foreground data. "
        "Computed from data if not set. "
        "[%default]")

    parser.add_option("--output-all-profiles", dest="output_all_profiles",
                      action="store_true",
                      help="keep individual profiles for each "
                      "transcript and output. "
                      "[%default]")

    parser.add_option("--counts-tsv-file", dest="input_filename_counts",
                      type="string",
                      help="filename with count data for each transcript. "
                      "Use this instead "
                      "of recomputing the profile. Useful for plotting the "
                      "meta-gene profile "
                      "from previously computed counts "
                      "[%default]")

    parser.add_option(
        "--background-region-bins",
        dest="background_region_bins",
        type="int",
        help="number of bins on either end of the profile "
        "to be considered for background meta-gene normalization "
        "[%default]")

    parser.set_defaults(
        remove_rna=False,
        ignore_pairs=False,
        force_output=False,
        bin_size=10,
        extends=[],
        shifts=[],
        sort=[],
        reporter="transcript",
        resolution_cds=1000,
        resolution_introns=1000,
        # 3kb is a good balance of seeing long enough 3 prime bias and not omit
        # too many genes. Tim 31th Aug 2013
        resolution_exons_absolute_distance_topolya=3000,
        # introns is only for assess the noise level, thus do ont need a long
        # region, a long region has the side effect of omit more genes. Tim
        # 31th Aug 2013
        resolution_introns_absolute_distance_topolya=500,
        # extension can simply just be the same as resolution
        extension_exons_absolute_distance_topolya=3000,
        extension_introns_absolute_distance_topolya=500,
        resolution_upstream_utr=1000,
        resolution_downstream_utr=1000,
        resolution_upstream=1000,
        resolution_downstream=1000,
        resolution_first=1000,
        resolution_last=1000,
        # mean length of transcripts: about 2.5 kb
        extension_upstream=2500,
        extension_downstream=2500,
        extension_inward=3000,
        extension_outward=3000,
        plot=True,
        methods=[],
        infiles=[],
        controlfiles=[],
        gtffile=None,
        profile_normalizations=[],
        transcript_normalization=None,
        scale_flanks=0,
        merge_pairs=False,
        min_insert_size=0,
        max_insert_size=1000,
        base_accuracy=False,
        matrix_format="single",
        control_factor=None,
        output_all_profiles=False,
        background_region_bins=10,
        input_filename_counts=None,
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv, add_output_options=True)

    # Keep for backwards compatability
    if len(args) == 2:
        infile, gtf = args
        options.infiles.append(infile)
        options.gtffile = gtf

    if not options.gtffile:
        raise ValueError("no GTF file specified")

    if options.gtffile == "-":
        options.gtffile = options.stdin
    else:
        options.gtffile = IOTools.openFile(options.gtffile)

    if len(options.infiles) == 0:
        raise ValueError("no bam/wig/bed files specified")

    for methodsRequiresBaseAccuracy in [
            "geneprofilewithintrons",
            "geneprofileabsolutedistancefromthreeprimeend",
    ]:
        # If you implemented any methods that you do not want the
        # spliced out introns or exons appear to be covered by
        # non-existent reads, it is better you let those methods imply
        # --base-accurarcy by add them here.
        if methodsRequiresBaseAccuracy in options.methods:
            options.base_accuracy = True

    if options.reporter == "gene":
        gtf_iterator = GTF.flat_gene_iterator(GTF.iterator(options.gtffile))
    elif options.reporter == "transcript":
        gtf_iterator = GTF.transcript_iterator(GTF.iterator(options.gtffile))

    # Select rangecounter based on file type
    if len(options.infiles) > 0:
        if options.infiles[0].endswith(".bam"):
            bamfiles = [pysam.AlignmentFile(x, "rb") for x in options.infiles]

            if options.controlfiles:
                controlfiles = [pysam.AlignmentFile(x, "rb")
                                for x in options.controlfiles]
            else:
                controlfiles = None

            format = "bam"
            if options.merge_pairs:
                range_counter = _bam2geneprofile.RangeCounterBAM(
                    bamfiles,
                    shifts=options.shifts,
                    extends=options.extends,
                    merge_pairs=options.merge_pairs,
                    min_insert_size=options.min_insert_size,
                    max_insert_size=options.max_insert_size,
                    controfiles=controlfiles,
                    control_factor=options.control_factor)

            elif options.shifts or options.extends:
                range_counter = _bam2geneprofile.RangeCounterBAM(
                    bamfiles,
                    shifts=options.shifts,
                    extends=options.extends,
                    controlfiles=controlfiles,
                    control_factor=options.control_factor)

            elif options.base_accuracy:
                range_counter = _bam2geneprofile.RangeCounterBAMBaseAccuracy(
                    bamfiles,
                    controlfiles=controlfiles,
                    control_factor=options.control_factor)
            else:
                range_counter = _bam2geneprofile.RangeCounterBAM(
                    bamfiles,
                    controlfiles=controlfiles,
                    control_factor=options.control_factor)

        elif options.infiles[0].endswith(".bed.gz"):
            bedfiles = [pysam.Tabixfile(x) for x in options.infiles]

            if options.controlfiles:
                controlfiles = [pysam.Tabixfile(x)
                                for x in options.controlfiles]
            else:
                controlfiles = None

            range_counter = _bam2geneprofile.RangeCounterBed(
                bedfiles,
                controlfiles=controlfiles,
                control_factor=options.control_factor)

        elif options.infiles[0].endswith(".bw"):
            wigfiles = [BigWigFile(file=open(x)) for x in options.infiles]
            range_counter = _bam2geneprofile.RangeCounterBigWig(wigfiles)

        else:
            raise NotImplementedError(
                "can't determine file type for %s" % str(options.infiles))

    counters = []
    for method in options.methods:
        if method == "utrprofile":
            counters.append(_bam2geneprofile.UTRCounter(
                range_counter,
                options.resolution_upstream,
                options.resolution_upstream_utr,
                options.resolution_cds,
                options.resolution_downstream_utr,
                options.resolution_downstream,
                options.extension_upstream,
                options.extension_downstream,
            ))

        elif method == "geneprofile":
            counters.append(_bam2geneprofile.GeneCounter(
                range_counter,
                options.resolution_upstream,
                options.resolution_cds,
                options.resolution_downstream,
                options.extension_upstream,
                options.extension_downstream,
                options.scale_flanks))

        elif method == "geneprofilewithintrons":
            counters.append(_bam2geneprofile.GeneCounterWithIntrons(
                range_counter,
                options.resolution_upstream,
                options.resolution_cds,
                options.resolution_introns,
                options.resolution_downstream,
                options.extension_upstream,
                options.extension_downstream,
                options.scale_flanks))

        elif method == "geneprofileabsolutedistancefromthreeprimeend":
            # options.extension_exons_absolute_distance_tostartsite,
            # options.extension_introns_absolute_distance_tostartsite,
            # Tim 31th Aug 2013: a possible feature for future,  if five prime
            # bias is of your interest.
            # (you need to create another class). It is not very difficult to
            # derive from this class, but is not implemented yet
            # This future feature is slightly different the TSS profile
            # already implemented, because in this future feature introns are
            # skipped,
            counters.append(
                _bam2geneprofile.GeneCounterAbsoluteDistanceFromThreePrimeEnd(
                    range_counter, options.resolution_upstream,
                    options.resolution_downstream,
                    options.resolution_exons_absolute_distance_topolya,
                    options.resolution_introns_absolute_distance_topolya,
                    options.extension_upstream,
                    options.extension_downstream,
                    options.extension_exons_absolute_distance_topolya,
                    options.extension_introns_absolute_distance_topolya,
                    options.scale_flanks))

        elif method == "tssprofile":
            counters.append(_bam2geneprofile.TSSCounter(
                range_counter,
                options.extension_outward,
                options.extension_inward))

        elif method == "intervalprofile":
            counters.append(_bam2geneprofile.RegionCounter(
                range_counter,
                options.resolution_upstream,
                options.resolution_cds,
                options.resolution_downstream,
                options.extension_upstream,
                options.extension_downstream))

        elif method == "midpointprofile":
            counters.append(_bam2geneprofile.MidpointCounter(
                range_counter,
                options.resolution_upstream,
                options.resolution_downstream,
                options.extension_upstream,
                options.extension_downstream))

        # add new method to split 1st and last exons out
        # requires a representative transcript for reach gene
        # gtf should be sorted gene-position
        elif method == "separateexonprofile":
            counters.append(_bam2geneprofile.SeparateExonCounter(
                range_counter,
                options.resolution_upstream,
                options.resolution_first,
                options.resolution_last,
                options.resolution_cds,
                options.resolution_downstream,
                options.extension_upstream,
                options.extension_downstream))

        elif method == "separateexonprofilewithintrons":
            counters.append(_bam2geneprofile.SeparateExonWithIntronCounter(
                range_counter,
                options.resolution_upstream,
                options.resolution_first,
                options.resolution_last,
                options.resolution_cds,
                options.resolution_introns,
                options.resolution_downstream,
                options.extension_upstream,
                options.extension_downstream))

    # set normalization
    for c in counters:
        c.setNormalization(options.transcript_normalization)
        if options.output_all_profiles:
            c.setOutputProfiles(IOTools.openFile(E.getOutputFile(c.name) +
                                                 ".profiles.tsv.gz", "w"))

    if options.input_filename_counts:
        # read counts from file
        E.info("reading counts from %s" % options.input_filename_counts)
        all_counts = pandas.read_csv(
            IOTools.openFile(options.input_filename_counts),
            sep='\t', header=0, index_col=0)

        if len(counters) != 1:
            raise NotImplementedError(
                'counting from matrix only implemented for 1 counter.')
        # build counter based on reference counter
        counter = _bam2geneprofile.UnsegmentedCounter(counters[0])
        counters = [counter]
        _bam2geneprofile.countFromCounts(counters, all_counts)

    else:
        E.info("starting counting with %i counters" % len(counters))
        feature_names = _bam2geneprofile.countFromGTF(counters,
                                                      gtf_iterator)

    # output matrices
    if not options.profile_normalizations:
        options.profile_normalizations.append("none")
    elif "all" in options.profile_normalizations:
        options.profile_normalizations = ["none",
                                          "area",
                                          "counts",
                                          "background"]

    for method, counter in zip(options.methods, counters):
        profiles = []
        for norm in options.profile_normalizations:
            # build matrix, apply normalization
            profile = counter.getProfile(
                normalize=norm,
                background_region_bins=options.background_region_bins)
            profiles.append(profile)

        for x in range(1, len(profiles)):
            assert profiles[0].shape == profiles[x].shape

        # build a single matrix of all profiles for output
        matrix = numpy.concatenate(profiles)
        matrix.shape = len(profiles), len(profiles[0])
        matrix = matrix.transpose()

        with IOTools.openFile(E.getOutputFile(counter.name) +
                              ".matrix.tsv.gz", "w") as outfile:
            outfile.write("bin\tregion\tregion_bin\t%s\n" % "\t".join(
                options.profile_normalizations))
            fields = []
            bins = []
            for field, nbins in zip(counter.fields, counter.nbins):
                fields.extend([field] * nbins)
                bins.extend(list(range(nbins)))

            for row, cols in enumerate(zip(fields, bins, matrix)):
                outfile.write("%i\t%s\t" %
                              (row, "\t".join([str(x) for x in cols[:-1]])))
                outfile.write("%s\n" %
                              ("\t".join([str(x) for x in cols[-1]])))

        with IOTools.openFile(E.getOutputFile(counter.name) +
                              ".lengths.tsv.gz", "w") as outfile:
            counter.writeLengthStats(outfile)

        if options.output_all_profiles:
            counter.closeOutputProfiles()

    if options.plot:

        import matplotlib
        # avoid Tk or any X
        matplotlib.use("Agg")
        import matplotlib.pyplot as plt

        for method, counter in zip(options.methods, counters):

            if method in ("geneprofile",
                          "geneprofilewithintrons",
                          "geneprofileabsolutedistancefromthreeprimeend",
                          "utrprofile",
                          "intervalprofile",
                          "separateexonprofile",
                          "separateexonprofilewithintrons"):

                plt.figure()
                plt.subplots_adjust(wspace=0.05)
                max_scale = max([max(x) for x in counter.aggregate_counts])

                for x, counts in enumerate(counter.aggregate_counts):
                    plt.subplot(6, 1, x + 1)
                    plt.plot(list(range(len(counts))), counts)
                    plt.title(counter.fields[x])
                    plt.ylim(0, max_scale)

                figname = counter.name + ".full"

                fn = E.getOutputFile(figname) + ".png"
                plt.savefig(os.path.expanduser(fn))

                plt.figure()

                points = []
                cuts = []
                for x, counts in enumerate(counter.aggregate_counts):
                    points.extend(counts)
                    cuts.append(len(counts))

                plt.plot(list(range(len(points))), points)

                xx, xxx = 0, []
                for x in cuts:
                    xxx.append(xx + x // 2)
                    xx += x
                    plt.axvline(xx,
                                color="r",
                                ls="--")

                plt.xticks(xxx, counter.fields)

                figname = counter.name + ".detail"

                fn = E.getOutputFile(figname) + ".png"
                plt.savefig(os.path.expanduser(fn))

            elif method == "tssprofile":

                plt.figure()
                plt.subplot(1, 3, 1)
                plt.plot(list(range(-options.extension_outward,
                                    options.extension_inward)),
                         counter.aggregate_counts[0])
                plt.title(counter.fields[0])
                plt.subplot(1, 3, 2)
                plt.plot(list(range(-options.extension_inward,
                                    options.extension_outward)),
                         counter.aggregate_counts[1])
                plt.title(counter.fields[1])
                plt.subplot(1, 3, 3)
                plt.title("combined")
                plt.plot(list(range(-options.extension_outward,
                                    options.extension_inward)),
                         counter.aggregate_counts[0])
                plt.plot(list(range(-options.extension_inward,
                                    options.extension_outward)),
                         counter.aggregate_counts[1])
                plt.legend(counter.fields[:2])

                fn = E.getOutputFile(counter.name) + ".png"
                plt.savefig(os.path.expanduser(fn))

            elif method == "midpointprofile":

                plt.figure()
                plt.plot(numpy.arange(-options.resolution_upstream, 0),
                         counter.aggregate_counts[0])
                plt.plot(numpy.arange(0, options.resolution_downstream),
                         counter.aggregate_counts[1])

                fn = E.getOutputFile(counter.name) + ".png"
                plt.savefig(os.path.expanduser(fn))

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''diff_gtf.py - compute overlap between multiple gtf files
===========================================================

:Tags: Genomics Intervals Genesets GTF Comparison

Purpose
-------

This script compares multiple set of gtf files. It computes
the overlap between bases, exons and genes between each pair
of gtf files.

If results from a previous run are present, existing
pairs are not re-computed but simply echoed.

The output is a tab-separated table with counts for each pair
of files being compared. The fields are:

+--------------+----------------------------------+
|*Column*      |*Content*                         |
+--------------+----------------------------------+
|set           |Name of the set                   |
+--------------+----------------------------------+
|ngenes_total  |number of genes in set            |
+--------------+----------------------------------+
|ngenes_ovl    |number of genes overlapping       |
+--------------+----------------------------------+
|ngenes_unique |number of unique genes            |
+--------------+----------------------------------+
|nexons_total  |number of exons in set            |
+--------------+----------------------------------+
|nexons_ovl    |number of exons overlapping       |
+--------------+----------------------------------+
|nexons_unique |number of unique exons            |
+--------------+----------------------------------+
|nbases_total  |number of bases in gene set       |
+--------------+----------------------------------+
|nbases_ovl    |number of bases overlapping       |
+--------------+----------------------------------+
|nbases_unique |number of unique bases            |
+--------------+----------------------------------+

Each of these fields will appear twice, once for each of the pair of files.
Hence ngenes_unique1 will be the number of genes in set1 that have no exons
that overlap with any exons in set2, and vice versa for ngenes_unique2. And
on for each field in the table above. This makes a total of 9*2=18 fields
containing counts, each starting with an n.

A further set of 18 fields each start with a ``p`` and are the corresponding
percentage values.

Options
-------

-s, --ignore-strand
    Ignore strand infomation so that bases overlap even if exons/genes are
    on different strands

-u, --update=FILENAME
    Read in previous results from FILENAME and only output comparisons that
    are missing.

-p, --pattern-identifier=PATTERN
    Provide a regular expression pattern for converting a filename into a
    set name for the output. The regular expression should capture at least
    one group. That group will be used to identify that file in the output
    table (see examples)


Examples
--------

For example if we have two gtf_files that look like::

   first_set_of_genes.gtf:
   1	protein_coding	exon	1	10	.	+	.	gene_id "1"; transcript_id "1"
   1	protein_coding	exon	20	30	.	+	.	gene_id "1"; transcript_id "1"

   second_set_of_genes.gtf:
   1	protein_coding	exon	25	35	.	+	.	gene_id "1"; transcript_id "1"
   2	protein_coding	exon	100	200	.	+	.	gene_id "2"; transcript_id "3"

Then the command::

   python diff_gtf.py *.gtf --pattern-identifier='(.+)_of_genes.gtf' > out.tsv

would produce an output file that has a single row with set1 being "second_set"
and set2 being "first_set" (these are extracted using that --pattern-identifier
option). It will report that set1 contains 2 genes and set2 1 gene. That for
each set one of these genes overlaps with the other set. For set1 it will
report that 1 gene is unique and that no genes are unique for set2 and so on
for exons and bases.

If we want to add a third file to the comparison,
"third_set_of_genes.gtf", we don't need to redo the comparison between
first_set_of_genes.gtf and second_set_of_genes.gtf::

   python diff_gtf.py --update=out.tsv *.gtf.gz > new.tsv

This will output a table with a row for third_set vs second_set and
third_set vs second_set, along with the comparison of first_set and
second_set that will simply be copied from the previous results. It is
important to include all files on the command line that are to be
output. Any comparisons in ``out.tsv`` that correspond to files that
are not given on the command line will not be output.

Usage
-----

::
   cgat diff_gtf.py GTF GTF [GTF [GTF [...]]] [OPTIONS]
   cgat diff_gtf GTF1 --update=OUTFILE [OPTIONS]

where GTF is a gtf or compressed gtf formated file and OUTFILE is the results
from a previous run.  At least two must be provided unless --update is present.

Type::

   python diff_gtf.py --help

for command line help.

Command line options
--------------------

'''

import sys
import re
import numpy

import CGAT.Experiment as E
import CGAT.GTF as GTF
import CGAT.IOTools as IOTools
import CGAT.NCL as NCL


class Counter:

    mPercentFormat = "%5.2f"

    def __init__(self):
        pass

    def getHeader(self):
        h = []
        for a in ("genes", "exons", "bases"):
            for b in ("total", "ovl", "unique"):
                for c in ("1", "2"):
                    h.append("n" + a + "_" + b + c)
        for a in ("genes", "exons", "bases"):
            for b in ("ovl", "unique"):
                for c in ("1", "2"):
                    h.append("p" + a + "_" + b + c)

        return "\t".join(h)

    @E.cachedmethod
    def buildIndex(self, filename):
        """read and index."""

        idx = {}
        infile = IOTools.openFile(filename, "r")
        for e in GTF.readFromFile(infile):
            if e.contig not in idx:
                idx[e.contig] = NCL.NCLSimple()
            idx[e.contig].add(e.start, e.end)
        infile.close()
        return idx

    def _count(self, filename, idx):

        overlapping_genes = set()
        genes = set()
        # iterate over exons
        infile = IOTools.openFile(filename, "r")
        it = GTF.iterator(infile)

        nexons, nexons_overlapping = 0, 0
        nbases, nbases_overlapping = 0, 0
        for this in it:
            nexons += 1
            nbases += this.end - this.start
            genes.add(this.gene_id)

            try:
                intervals = list(idx[this.contig].find(this.start, this.end))
            except KeyError:
                continue

            if len(intervals) == 0:
                continue

            overlapping_genes.add(this.gene_id)
            nexons_overlapping += 1
            start, end = this.start, this.end
            counts = numpy.zeros(end - start, numpy.int)
            for other_start, other_end, other_value in intervals:
                for x in range(max(start, other_start) - start, min(end, other_end) - start):
                    counts[x] += 1
            nbases_overlapping += sum([1 for x in counts if x > 0])

        infile.close()

        return len(genes), len(overlapping_genes), nexons, nexons_overlapping, nbases, nbases_overlapping

    def count(self, filename1, filename2):
        """count overlap between two gtf files."""

        E.info("counting started for %s versus %s" % (filename1, filename2))

        idx2 = self.buildIndex(filename2)

        (self.mGenes1, self.mGenesOverlapping1,
         self.mExons1, self.mExonsOverlapping1,
         self.mBases1, self.mBasesOverlapping1 ) = \
            self._count(filename1, idx2)

        self.mGenesUnique1 = self.mGenes1 - self.mGenesOverlapping1
        self.mExonsUnique1 = self.mExons1 - self.mExonsOverlapping1
        self.mBasesUnique1 = self.mBases1 - self.mBasesOverlapping1

        idx1 = self.buildIndex(filename1)

        (self.mGenes2, self.mGenesOverlapping2,
         self.mExons2, self.mExonsOverlapping2,
         self.mBases2, self.mBasesOverlapping2 ) = \
            self._count(filename2, idx1)

        self.mGenesUnique2 = self.mGenes2 - self.mGenesOverlapping2
        self.mExonsUnique2 = self.mExons2 - self.mExonsOverlapping2
        self.mBasesUnique2 = self.mBases2 - self.mBasesOverlapping2

    def __str__(self):

        return "\t".join(map(str, (
            self.mGenes1, self.mGenes2,
            self.mGenesOverlapping1, self.mGenesOverlapping2,
            self.mGenesUnique1, self.mGenesUnique2,
            self.mExons1, self.mExons2,
            self.mExonsOverlapping1, self.mExonsOverlapping2,
            self.mExonsUnique1, self.mExonsUnique2,
            self.mBases1, self.mBases2,
            self.mBasesOverlapping1, self.mBasesOverlapping2,
            self.mBasesUnique1, self.mBasesUnique2 ) ) ) + "\t" +\
            "\t".join([IOTools.prettyPercent(*x) for x in (
                (self.mGenesOverlapping1, self.mGenes1),
                (self.mGenesOverlapping2, self.mGenes2),
                (self.mGenesUnique1, self.mGenes1),
                (self.mGenesUnique2, self.mGenes2),
                (self.mExonsOverlapping1, self.mExons1),
                (self.mExonsOverlapping2, self.mExons2),
                (self.mExonsUnique1, self.mExons1),
                (self.mExonsUnique2, self.mExons2),
                (self.mBasesOverlapping1, self.mBases1),
                (self.mBasesOverlapping2, self.mBases2),
                (self.mBasesUnique1, self.mBases1),
                (self.mBasesUnique2, self.mBases2))])


class CounterGenes(Counter):

    """output only genes."""

    mSeparator = ";"

    def __init__(self, *args, **kwargs):
        Counter.__init__(self, *args, **kwargs)

    def getHeader(self):
        h = ["ngenes1", "ngenes2", "novl1", "novl2", "nuniq1",
             "nuniq2", "ovl1", "ovl2", "uniq1", "uniq2"]
        return "\t".join(h)

    def _count(self, filename, idx):

        overlapping_genes = set()
        genes = set()
        # iterate over exons
        infile = IOTools.openFile(filename, "r")
        it = GTF.iterator(infile)

        for this in it:
            genes.add(this.gene_id)

            try:
                intervals = idx[this.contig].find(this.start, this.end)
            except KeyError:
                continue

            if len(intervals) == 0:
                continue
	
            overlapping_genes.add(this.gene_id)

        infile.close()

        return genes, overlapping_genes

    def count(self, filename1, filename2):
        """count overlap between two gtf files."""

        E.info("counting started for %s versus %s" % (filename1, filename2))

        idx2 = self.buildIndex(filename2)

        (self.mGenes1, self.mGenesOverlapping1) = self._count(filename1, idx2)

        idx1 = self.buildIndex(filename1)
        (self.mGenes2, self.mGenesOverlapping2) = self._count(filename2, idx1)

    def __str__(self):

        uniq1 = self.mGenes1.difference(self.mGenesOverlapping)
        uniq2 = self.mGenes2.difference(self.mGenesOverlapping)

        return "\t".join(map(str, (
            len(self.mGenes1),
            len(self.mGenes2),
            len(self.mGenesOverlapping1),
            len(self.mGenesOverlapping2),
            len(uniq1),
            len(uniq2),
            self.mSeparator.join(self.mGenesOverlapping1),
            self.mSeparator.join(self.mGenesOverlapping2),
            self.mSeparator.join(uniq1),
            self.mSeparator.join(uniq2))))


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option("-s", "--ignore-strand", dest="ignore_strand",
                      action="store_true",
                      help="ignore strand information [default=%default].")

    parser.add_option(
        "-u", "--update", dest="filename_update", type="string",
        help="if filename is given, previous results will be read"
        "from there and only changed sets will be computed "
        "[default=%default].")

    parser.add_option(
        "-p", "--pattern-identifier", dest="pattern_id", type="string",
        help="pattern to convert a filename to an id"
        "[default=%default].")

    parser.add_option(
        "-g", "--output-only-genes", dest="output_only_genes",
        action="store_true",
        help="only output gene stats (includes gene lists)"
        " [default=%default].")

    parser.set_defaults(
        ignore_strand=False,
        filename_update=None,
        pattern_id="(.*).gtf",
        output_only_genes=False,
    )

    (options, args) = E.Start(parser)

    if len(args) < 2:
        print(USAGE)
        raise ValueError("at least two arguments required")

    if options.filename_update:
        infile = IOTools.openFile(options.filename_update, "r")
        previous_results = {}
        for line in infile:
            if line.startswith("#"):
                continue
            if line.startswith("set1"):
                continue
            data = line[:-1].split("\t")
            set1, set2 = data[0], data[1]

            if set1 not in previous_results:
                previous_results[set1] = {}
            if set2 not in previous_results:
                previous_results[set2] = {}

            previous_results[set1][set2] = "\t".join(data[2:])
            rev = [(data[x + 1], data[x]) for x in range(2, len(data), 2)]
            previous_results[set2][set1] = "\t".join(IOTools.flatten(rev))
    else:
        previous_results = {}

    if options.output_only_genes:
        counter = CounterGenes()
    else:
        counter = Counter()

    options.stdout.write("set1\tset2\t%s\n" % counter.getHeader())

    pattern_id = re.compile(options.pattern_id)

    def getTitle(x):
        try:
            return pattern_id.search(x).groups()[0]
        except AttributeError:
            return x

    ncomputed, nupdated = 0, 0
    for x in range(len(args)):
        title1 = getTitle(args[x])
        for y in range(0, x):
            title2 = getTitle(args[y])
            if previous_results:
                try:
                    prev = previous_results[title1][title2]
                except KeyError:
                    pass
                else:
                    options.stdout.write(
                        "%s\t%s\t%s\n" % ((title1, title2, prev)))
                    nupdated += 1
                    continue

            counter.count(args[x], args[y])
            options.stdout.write(
                "%s\t%s\t%s\n" % ((title1, title2, str(counter))))
            ncomputed += 1

    E.info("nupdated=%i, ncomputed=%i" % (nupdated, ncomputed))

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''CBioPortal.py - Interface with the Sloan-Kettering cBioPortal webservice
========================================================================

The Sloan Kettering cBioPortal webservice provides access to a
database of results of genomics experiments on various cancers. The
database is organised into studies, each study contains a number of
case lists, where each list contains the ids of a set of patients, and
genetic profiles, each of which represents an assay conducted on the
patients in the case list as part of the study.

The main class here is the CBioPortal class representing a connection
to the cBioPortal Database. Query's are represented as methods of the
class. Study ids or names or case lists can be provided to the
constructor to the object, via the setDefaultStudy and
setDefaultCaseList methods or to the indevidual query methods. Where
ever possible the validity of parameters is checked *before* the query
is executed.

Whenever a query requires a genetic profile id or a list of such ids,
but none are given, the list of all profiles for which the
show_in_analysis flag is set will be used.

All of the commands provided in the webservice are implemented here
and as far as possible the name, syntax and paramter names of the
query are identical to the raw commands to the webservice. These
queries are:

* getCancerStudies,
* getCaseLists,
* getProfileData,
* getMutationData,
* getClinicalData,
* getProteinArrayInfo,
* getProteinArrayData,
* getLink,
* getOncoprintHTML.

In addition two new queries are implememented that are not part of the
webservice:

* getPercentAltered and
* getTotalAltered

These emulate the function of the website where the percent of cases
that show any alteration for the gene and profiles given are returned
(getPercentAltered, or the percent of cases that show an alteration in
any of the genes (getTotalAltered) is returned.

examples::

   gene_list = [ "TP53",
   "BCL2",
   "MYC"  ]
   portal = CBioPortal()
   portal.setDefaultStudy(study = "prad_mskcc")
   portal.setDefaultCaseList(case_set_id = "prad_all_complete")
   portal.getPercentAltered(gene_list = gene_list)

or more tersely::

   portal.CBioProtal()
   portal.getPercentAltered(study = "prad_mskcc", case_set_id = "prad_all_complete",
                            gene_list = ["TP53","BCL2","MYC"],
                            genetic_profile_id =["prad_mskcc_mrna"])

Any warnings returned by the query are stored in CBioPortal.last_warnings.

Query's that would give too long an URL are split into smaller querys
and the results combined transparently.

A commandline interface is provided for convenience, syntax::

   python CBioPortal.py [options] command(s)

Reference
---------

'''
from future.moves.urllib.request import urlopen
import re
import optparse
import sys
from CGAT import IOTools as IOTools
from collections import OrderedDict as odict
from CGAT import Experiment as E


class CBioPortal():
    """connect to the cBioPortal Database.

    If no url is specified the default url is used. A list of of valid
    study ids is retrieved from the database. This both confirms that
    the datavase is reachable, and provides cached checking for the
    ids provided. If a study or study name is provided then this is
    set as the defualt study for this session and the details of the
    availible profiles and cases is retrieved.  'Study' is the study
    id. If both study and study_name are specified then the study id
    is used.
    """

    url = "http://www.cbioportal.org/public-portal/webservice.do"

    def __init__(self, url=None, study=None, study_name=None,
                 case_list_id=None):

        if url:
            self.url = url

        self.getCancerStudies()
        self.study = None
        self.case_list = None

        if study:

            if study in self._valid_study_ids:
                self.study = study

                self.profiles = self.getGeneticProfiles(study)
                self.cases = self.getCaseLists(study)
            else:
                raise ValueError("%s is not a valid study id" % study)

        elif study_name:
            if study_name in [x['name'] for x in cancer_studies]:
                study = [x['cancer_study_id']
                         for x in cancer_studies if x['name'] == study_name][0]
                self.study = study
                self.profiles = self.getGeneticProfiles(study)
                self.cases = self.getCaseLists(study)

            else:
                raise ValueError("%s is not a valid study name" % study_name)

        if case_list_id:
            self.setDefaultCaseList(case_list_id)

    def _executeQuery(self, command, args=None):
        """execute the provided command on the database.

        args are specified as a dictionary.  error checking on the
        results is performed here and the returned value is a list of
        lists with each list representing a row of the returned table.
        """
        try:
            argument_string = "&".join(["%s=%s" % (x, args[x]) for x in args])
            command_string = "&".join([command, argument_string])
        except TypeError:
            command_string = command

        query = "%s?cmd=%s" % (self.url, command_string)

        if len(query) > 2048:
            if "gene_list" in args:
                genes = args['gene_list'].split(",")

                if len(genes) < 2:
                    raise ValueError("Request too long")

                args['gene_list'] = ",".join(genes[(len(genes) / 2):])
                query1 = self._executeQuery(command, args)
                warnings = self.last_warnings

                args['gene_list'] = ",".join(genes[:(len(genes) / 2)])
                query2 = self._executeQuery(command, args)
                self.last_warnings = warnings + self.last_warnings

                return query1 + query2

        data = urlopen(query)

        line = data.readline()
        self.last_query = query
        self.last_status = line
        self.last_warnings = []
        self.last_header = [self.last_status]
        return_table = []

        while re.match("^#", line):

            if re.match("^# Warning: (.+)", line):
                self.last_warnings.append(
                    re.match("^# Warning: (.+)", line).groups(1)[0])
                self.last_header.append(line)
                line = data.readline()
                continue

            elif re.match("^#", line):
                self.last_header.append(line)
                line = data.readline()
                continue

        if re.match("^Error: (.+)", line):
            self.last_header.append(line)
            raise CDGSError(re.match("^Error: (.+)", line).groups(1)[0], query)
        line = line.strip()
        headers = line.split("\t")

        for line in data:
            if re.match("^# Warning: (.+)", line):
                self.last_warnings.append(
                    re.match("^# Warning: (.+)", line).groups(1)[0])
                self.last_header.append(line)
                line = data.readline()
                continue
            line = line.strip()
            return_table.append(odict(list(zip(headers, line.split("\t")))))

        return return_table

    def _getStudyId(self, study, study_name):

        if study:
            if study in self._valid_study_ids:
                return study
            else:
                raise ValueError("%s is not a valid study id" % study)

        elif study_name:
            name_lookup = [x['study']
                           for x in self.cancer_studies if x['name'] == study_name]
            if len(name_lookup) == 1:
                return name_lookup[0]
            else:
                raise ValueError(
                    "Cannot find study id for study '%s'" % study_name)

        elif self.study:
            return self.study

        else:
            return None

    def getCancerStudies(self):
        """Fetches the list of cancer studies currently in the database.

        Returns list of dictionaries with three entries
        'cancer_study_id','name' and 'description'.  Also caches this
        data to verify the validity of later calls
        """
        cancer_studies = self._executeQuery("getCancerStudies")
        self.cancer_studies = cancer_studies
        self._valid_study_ids = [x['cancer_study_id'] for x in cancer_studies]

        return self.cancer_studies

    def getGeneticProfiles(self, study=None, study_name=None):
        """Fetches the valid genetic profiles for a particular study.

        study is the study id.  If both study and study_name are
        specified, study is used. If neither study nor study name is
        specified then the default study is used if set, if not a
        value error is raised.
        Returns a list of dictionaries
        """

        study = self._getStudyId(study, study_name)

        if not study:
            raise ValueError("A study must be specified if no default is set")

        genetic_studies = self._executeQuery(command="getGeneticProfiles",
                                             args=dict(cancer_study_id=study))
        return genetic_studies

    def getCaseLists(self, study=None, study_name=None):
        """Retrieves meta-data regarding all case lists stored about a
        specific cancer study.

        For example, a within a particular study, only some cases may
        have sequence data, and another subset of cases may have been
        sequenced and treated with a specific therapeutic
        protocol. Multiple case lists may be associated with each
        cancer study, and this method enables you to retrieve
        meta-data regarding all of these case lists.

        Data is returned as a list of dictionaries with the following
        entries:

         * case_list_id: a unique ID used to identify the case list ID
           in subsequent interface calls. This is a human readable
           ID. For example, "gbm_all" identifies all cases profiles in
           the TCGA GBM study.

         * case_list_name: short name for the case list.

         * case_list_description: short description of the case list.

         * cancer_study_id: cancer study ID tied to this genetic
           profile. Will match the input cancer_study_id.

         * case_ids: space delimited list of all case IDs that make up
           this case list.
        """

        study = self._getStudyId(study, study_name)

        if not study:
            raise ValueError("A study must be specified if no default is set")

        case_lists = self._executeQuery(
            command="getCaseLists",
            args=dict
            ({'cancer_study_id': study}))
        return case_lists

    def getProfileData(self, gene_list, case_set_id=None,
                       genetic_profile_id=None, study=None, study_name=None):
        """Retrieves genomic profile data for one or more genes.

        You can specify one gene and many profiles or one profile and
        many genes.  If you specify no genetic profiles then all
        genetic profiles for the specified or default study are used
        if the case_set_id is from that study otherwise a ValueError
        is raised.

        Return value depends on the parameters. If you specify a
        single genetic profile and multiple genes a list of ordered
        dictionaries with the following entries::

            gene_id: Entrez Gene ID
            common: HUGO Gene Symbol
            entries 3 - N: Data for each case

        If you specify multi genetic profiles and a single gene, a
        list of ordered dictoraries with the following entries is
        returned::

            genetic_profile_id: The Genetic Profile ID.
            alteration_type: The Genetic Alteration Type, e.g. MUTATION, MUTATION_EXTENDED, COPY_NUMBER_ALTERATION, or MRNA_EXPRESSION.
            gene_id: Entrez Gene ID.
            common: HUGO Gene Symbol.
            Columns 5 - N: Data for each case.

        """

        # Do some pre-query checking.

        if len(gene_list) > 1 and len(genetic_profile_id) > 1:
            raise ValueError(
                "%i genes and %i profiles specified\n.Please "
                "specify either one gene many profiles or one "
                "profile many genes" % (
                    len(gene_list), len(genetic_profile_id)))

        study_id = self._getStudyId(study, study_name)
        case_set_id = self._getCaseListId(case_set_id, study_id)

        genetic_profile_id = self._getAndCheckGeneticProfiles(
            genetic_profile_id=genetic_profile_id, study=study)

        gene_list = ",".join(gene_list)
        genetic_profile_id = ",".join(genetic_profile_id)

        profile_data = self._executeQuery(
            command="getProfileData",
            args={'case_set_id': case_set_id,
                  'genetic_profile_id': genetic_profile_id,
                  'gene_list': gene_list})

        return profile_data

    def getMutationData(self, gene_list, genetic_profile_id, case_set_id=None,
                        study=None, study_name=None):
        '''For data of type EXTENDED_MUTATION, you can request the full set of
        annotated extended mutation data.

        This enables you to, for example, determine which sequencing
        center sequenced the mutation, the amino acid change that
        results from the mutation, or gather links to predicted
        functional consequences of the mutation.

        Query Format


            case_set_id= [case set ID] (required)
            genetic_profile_id= [a single genetic profile IDs] (required).
            gene_list= [one or more genes, specified as HUGO Gene Symbols or
                 Entrez Gene IDs](required)

        Response Format

        A list of dictionaries with the following entires

            entrez_gene_id: Entrez Gene ID.
            gene_symbol: HUGO Gene Symbol.
            case_id: Case ID.
            sequencing_center: Sequencer Center responsible for identifying
                this mutation.
                               For example: broad.mit.edu.
            mutation_status: somatic or germline mutation status. all mutations
                         returned will be of type somatic.
            mutation_type: mutation type, such as nonsense, missense, or frameshift_ins.
            validation_status: validation status. Usually valid, invalid, or unknown.
            amino_acid_change: amino acid change resulting from the mutation.

            functional_impact_score: predicted functional impact score,
                 as predicted by: Mutation Assessor.
            xvar_link: Link to the Mutation Assessor web site.
            xvar_link_pdb: Link to the Protein Data Bank (PDB) View within
                       Mutation Assessor web site.
            xvar_link_msa: Link the Multiple Sequence Alignment (MSA) view
                 within the Mutation Assessor web site.
            chr: chromosome where mutation occurs.
            start_position: start position of mutation.
            end_position: end position of mutation.

        If a default study is set then a check will be performed to
        set if the supplied case id is from the specified study. The
        study can be over written using the study and study_name
        parameters

        '''

        study_id = self._getStudyId(study, study_name)

        case_set_id = self._getCaseListId(case_set_id, study_id)

        if study_id:
            if not study_id == self.study:
                profiles = getGeneticProfiles(study_id)
            else:
                profiles = self.profiles

            if genetic_profile_id not in proiles:
                raise ValueError(
                    "%s not a valid genetic profile for study %s" %
                    (genetic_profile_id, gene_id))
        genetic_profile_id = ",".join(genetic_profile_id)
        gene_list = ",".join(gene_list)
        mutation_data = self._executeQuery(
            command="getMutationData",
            args=dict({"case_set_id": case_set_id,
                       "genetic_profile_id": genetic_profile_id,
                       "gene_list": gene_list}))
        return mutation_data

    def getClinicalData(self, case_set_id=None, study=None, study_name=None):
        '''Retrieves overall survival, disease free survival and age at
        diagnosis for specified cases.

        Due to patient privacy restrictions, no other clinical data is
        available.

        Query Format
        ------------

            case_set_id= [case set ID] (required)

        Response Format
        ---------------

        A list of dictionaries with the following entries:

            case_id: Unique Case Identifier.
            overall_survival_months: Overall survival, in months.
            overall_survival_status: Overall survival status, usually
                 indicated as "LIVING" or "DECEASED".
            disease_free_survival_months: Disease free survival, in months.
            disease_free_survival_status: Disease free survival status,
                 usually indicated as "DiseaseFree" or "Recurred/Progressed".
            age_at_diagnosis: Age at diagnosis.

        If a study is specified or a defualt study is set, then the
        case_set_id will be tested to check if it exists for that
        study.

        '''

        study_id = self._getStudyId(study, study_name)

        case_set_id = self._getCaseListId(case_set_id, study_id)

        clincal_data = self._executeQuery(
            command="getClinicalData",
            args=dict({'case_set_id': case_set_id}))
        return clincal_data

    def getProteinArrayInfo(self, protein_array_type=None, gene_list=None,
                            study=None, study_name=None):
        '''Retrieves information on antibodies used by reverse-phase protein
        arrays (RPPA) to measure protein/phosphoprotein levels.

        Query Format
        ------------

            cancer_study_id= [cancer study ID] (required)
            protein_array_type= [protein_level or phosphorylation]
            gene_list= [one or more genes, specified as HUGO Gene
            Symbols or Entrez Gene IDs].

        Response Format
        ---------------

        A list of dictionaries with the following entires:

            ARRAY_ID: The protein array ID.
            ARRAY_TYPE: The protein array antibody type, i.e. protein_level
                or phosphorylation.
            GENE: The targeted gene name (HUGO gene symbol).
            RESIDUE: The targeted resdue(s).

        If no study is specified the default study is used. If that is
        not specified an error is raised.

        '''

        study = self._getStudyId(study, study_name)
        args = dict({'cancer_study_id': study})

        if gene_list:
            args['gene_list'] = ",".join(gene_list)

        if protein_array_type:
            args['protein_array_type'] = protein_array_type

        protein_array_info = self._executeQuery(
            command="getProteinArrayInfo", args=args)

        return protein_array_info

    def getProteinArrayData(self, protein_array_id=None, case_set_id=None,
                            array_info=0, study=None, study_name=None):
        '''Retrieves protein and/or phosphoprotein levels measured by
        reverse-phase protein arrays (RPPA).

        Query Format
        ------------

        case_set_id= [case set ID]
        protein_array_id= [one or more protein array IDs] as list.
        array_info= [1 or 0]. If 1, antibody information will also be exported.

        Response Format 1
        -----------------

        If the parameter of array_info is not specified or it is not
        1, returns a list of dictionaries with the following columns.

        ARRAY_ID: The protein array ID.
        Columns 2 - N: Data for each case.

        Response Format 2
        -----------------

        If the parameter of array_info is 1, you will receive a list
        of ordered dictionaries with the following entires:

        ARRAY_ID: The protein array ID.
        ARRAY_TYPE: The protein array antibody type, i.e. protein_level or
             phosphorylation.
        GENE: The targeted gene name (HUGO gene symbol).
        RESIDUE: The targeted resdue(s).
        Columns 5 - N: Data for each case.

        If the defualt study is set then the case_set_id will be
        check. The default study can be overidden using the study or
        study_name parameters.

        '''

        study_id = self._getStudyId(study, study_name)

        case_set_id = self._getCaseListId(case_set_id, study_id)
        args = dict({"case_set_id": case_set_id,
                     "array_info": array_info})

        if protein_array_id:
            args['protein_array_id'] = ",".join(protein_array_id)

        protein_array_data = self._executeQuery(
            command="getProteinArrayData", args=args)

        return protein_array_data

    def getLink(self, gene_list, study=None, study_name=None,  report="full"):
        '''return a perminant link to the cBioPortal report for the gene_list
            cancer_study_id=[cancer study ID] gene_list=[a comma
            separated list of HUGO gene symbols] (required)
            report=[report to display; can be one of: full (default),
            oncoprint_html]
        '''

        study = self._getStudyId(study, study_name)
        if not study:
            raise ValueError("Study must be specified")
        if report not in ["full", "oncoprint_html"]:
            raise ValueError("%s is not a valid report" % report)

        url = "/".join(self.url.split("/")[:-1])

        gene_list = ",".join(gene_list)
        return "%s/link.do?cancer_study_id=%s&gene_list=%s&report=%s" %\
            (url, study, gene_list, report)

    def getOncoprintHTML(self, gene_list, study=None, study_name=None):
        '''returns the HTML for the oncoprint report for the specified gene
        list and study'''

        url = "/".join(self.url.split("/")[0:-1])
        gene_list = ",".join(gene_list)
        command = "%s/link.do?cancer_study_id=%s&gene_list=%s&report=oncoprint_html" % (
            url, study, gene_list)
        return urlopen(command).read()

    def setDefaultStudy(self, study=None, study_name=None):
        '''sets a new study as the default study. Will check that the study
        id is valid'''
        study = self._getStudyId(study, study_name)
        self.study = study
        self.profiles = self.getGeneticProfiles()
        self.cases = self.getCaseLists()

        all_case_list = [x['case_list_id']
                         for x in self.cases
                         if x['case_list_name'] == "All Tumours"]

        if len(all_case_list) == 1:
            self.setDefaultCaseList(all_case_list)

    def setDefaultCaseList(self, case_set_id, study=None, study_name=None):
        '''set the default case list. If study is not specified the default
        study will be used.

        The study will be used to check that the case_set exists.
        '''

        study = self._getStudyId(study, study_name)
        case_list = self._getCaseListId(case_set_id, study=study)

        if not study == self.study:
            self.setDefaultStudy(study)

        self.case_list = case_list

    def _getCaseListId(self, case_set_id=None, study=None, strict=True):
        ''' checking is only done if study is specified or a default is set '''

        study_id = self._getStudyId(study, None)

        if case_set_id:
            if study_id:
                if not study_id == self.study:
                    case_lists = [x['case_list_id']
                                  for x in self.getCaseLists(study_id)]
                else:
                    case_lists = [x['case_list_id'] for x in self.cases]

                if case_set_id not in case_lists:
                    raise ValueError(
                        "%s is not a valid case list for study %s" % (case_set_id, study_id))
            return case_set_id

        else:
            if self.case_list:
                return self.case_list
            else:
                raise ValueError("No case_set_id provided and no default set")

    def _getAndCheckGeneticProfiles(self, genetic_profile_id=None, study=None):

        study_id = self._getStudyId(study, None)
        if not genetic_profile_id:
            if not study_id:
                raise ValueError(
                    "Either genetic_profile_id or study must be specified")

            if study_id == self.study:
                genetic_profile_id = [x['genetic_profile_id']
                                      for x in self.profiles
                                      if x['show_profile_in_analysis_tab'] == "true"]
            else:
                genetic_profile_id = [x['genetic_profile_id']
                                      for x in self.getGeneticProfiles(study_id)
                                      if x['show_profile_in_analysis_tab'] == "true"]

            return genetic_profile_id

        else:

            if not study_id:
                return genetic_profile_id

            if study_id == self.study:
                genetic_profile_id = [x['genetic_profile_id'] for x in self.profiles
                                      if x['genetic_profile_id'] in genetic_profile_id]
            else:
                genetic_profile_id = [x['genetic_profile_id'] for x in self.getGeneticProfiles(study_id)
                                      if x['genetic_profile_id'] in genetic_profile_id]

            if len(genetic_profile_id) == 0:
                raise ValueError("no valid genetic_profile_ids found")

            return genetic_profile_id

    def getPercentAltered(self, gene_list, study=None, study_name=None,
                          case_set_id=None, genetic_profile_id=None,
                          threshold=2):
        '''Get the percent of cases that have one or more of the specified
        alterations for each gene

        Query Format
        ------------

        study = [cancer_study_id] The study to use.

        study_name = [cancer_study_name] The name of the study to
                     use. If neither this nor study are specified,
                     then the default is used.

        case_set_id = [case_set_id] The case list to use. If not
                      specified, the default case list is used.

        gene_list = [one or more genes, specified as HUGO Gene Symobls
        or ENtrez Gene IDs] (require)

        genetic_profile_id = [one or more genetic profile IDs] If none
        specified all genetic profiles for the specified study are
        used..

        threhold = [z_score_threshold] the numeric threshold at which
        a mrna expression z-score is said to be significant.

        Response Format
        ---------------

        A list of dictionaries with the following entries
        gene_id: The Entrez Gene ID
        common: The Hugo Gene Symbol
        altered_in: The percent of cases in which the gene is altered

        One implementation note is that a guess must be made as to
        wether a returned profile value represents a alteration or
        not. Currently guesses are only made for copy number
        variation, mrna expression and mutionation

        '''

        study = self._getStudyId(study, study_name)
        case_set_id = self._getCaseListId(case_set_id, study)
        genetic_profile_id = self._getAndCheckGeneticProfiles(
            genetic_profile_id, study)

        if study and study == self.study:
            profiles = self.profiles
            case_list = [x['case_ids']
                         for x in self.cases
                         if x['case_list_id'] == case_set_id][0]

        else:
            profiles = self.getGeneticProfiles(study)
            case_list = [x['case_ids'] for x in self.getCaseLists(
                study) if x['case_list_id'] == case_set_id][0]

        return_table = []
        case_list = case_list.split(" ")

        data = []
        warnings = []
        for profile in genetic_profile_id:

            data.append(self.getProfileData(
                gene_list=gene_list, case_set_id=case_set_id,
                genetic_profile_id=[profile]))
            warnings.extend(self.last_warnings)

        # data[profile][gene][case]
        for gene in range(len(data[0])):
            cases_altered = 0.0

            geneProfile = [x[gene] for x in data]

            for case in (set(geneProfile[0]) - set(["gene_id", "common"])):
                if len([geneProfile[x][case] for x in range(len(geneProfile))
                        if self._guessAlteration(
                            geneProfile[x][case],
                            genetic_profile_id[x], profiles)]) > 0:

                    cases_altered += 1

            return_table.append(
                dict({'gene_id': geneProfile[0]['GENE_ID'],
                      'common': geneProfile[0]['COMMON'],
                      'altered_in': cases_altered * 100 / len(case_list)}))

        self.last_warnings = warnings
        return return_table

    def getTotalAltered(self, gene_list, study=None, study_name=None, case_set_id=None, genetic_profile_id=None, threshold=2):
        ''' Calculate the percent of cases in which any one of the specified genes are altered '''

        study = self._getStudyId(study, study_name)
        case_set_id = self._getCaseListId(case_set_id, study)
        genetic_profile_id = self._getAndCheckGeneticProfiles(
            genetic_profile_id, study)

        if study == self.study:
            profiles = self.profiles
            case_list = [x['case_ids']
                         for x in self.cases if x['case_list_id'] == case_set_id][0]
        else:
            profiles = self.getGeneticProfiles(study)
            case_list = [x['case_ids'] for x in self.getCaseLists(
                study) if x['case_list_id'] == case_set_id][0]
        case_list = case_list.split(" ")

        data = []

        # data[gene][profile][case]
        cases_altered = 0.0
        for gene in gene_list:
            data.append(self.getProfileData(
                gene_list=[gene], case_set_id=case_set_id,
                genetic_profile_id=genetic_profile_id))
            # catch special case where only a single
            # genetic_profile_id is passed in, so the query is single
            # gene and single profile and returns genes as columns
            # format.
            if len(genetic_profile_id) == 1 and len(data[0]) == 1:
                data[0][0]['GENETIC_PROFILE_ID'] = genetic_profile_id[0]

        for case_id in set(data[0][0]) - set(
                ["GENETIC_PROFILE_ID", "ALTERATION_TYPE",
                 "GENE_ID", "COMMON"]):

            case_altered = False
            for gene in data:

                altered = len([x for x in gene
                               if self._guessAlteration(x[case_id],
                                                        x['GENETIC_PROFILE_ID'],
                                                        profiles, threshold)])
                if (altered > 0):
                    case_altered = True

            if case_altered is True:
                cases_altered += 1.0

        return cases_altered * 100 / len(case_list)

    def _guessAlteration(self, value, genetic_profile_id, genetic_profiles,
                         threshold=2):

        alteration_type = [x['genetic_alteration_type']
                           for x in genetic_profiles
                           if x['genetic_profile_id'] == genetic_profile_id][0]

        # print alteration_type
        if alteration_type == "COPY_NUMBER_ALTERATION":
            if value == "0" or value == "-1" or value == "1":
                return False
            elif value == "NaN":
                return False
            else:
                return True

        elif alteration_type == "MRNA_EXPRESSION":
            if re.search("[^0-9\.\-]", value):
                # is character string
                return False
            elif re.search("[0-9]+\.[0-9]+", value):
                # is float
                value = float(value)
                if abs(value) > threshold:
                    return True
                else:
                    return False
            else:  # must be int?
                # print value
                if value == "0":
                    return False
                else:
                    return True

        elif alteration_type == "METHYLATION":
            return False

        elif (alteration_type == "MUTATION" or
              alteration_type == "MUTATION_EXTENDED"):
            if value == "NaN":
                return False
            else:
                return True
        else:
            return False


class CDGSError(Exception):

    '''exception that handles errors returned by querys in the database'''

    def __init__(self, error, request):
        self.error = error
        self.request = request

    def __str__(self):
        return "Request %s return error:\n%s" % (self.request, self.error)


def tableToString(intable):

    headers = "\t".join([x for x in intable[0]])

    line_list = []
    for line in intable:

        line_list.append("\t".join([str(line[x]) for x in line]))

    outTable = "\n".join(line_list)
    outTable = headers + "\n" + outTable

    return outTable


def main(argv=None):

    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option("-o", "--output_file", type="string", default=None,
                      help="[Optional] Filename to output results to. [default=STDOUT]")
    parser.add_option("-u", "--url", type="string", default="http://www.cbioportal.org/public-portal/webservice.do",
                      help="[Optional] Url to the cBioPortal webservice [default=%default]")

    cqueryopts = optparse.OptionGroup(
        parser, "Common parameters", "Common arguments to the query")
    cqueryopts.add_option("-s", "--study_id", dest="study_id", type="string", default=None,
                          help="[Required/OPtional]  cBioPortal ID for study [default=%default].\n This or study_name required for: getGeneticProfiles, getCaseLists, getProteinArrayInfo, getLink,getOncoprintHTML, getPercentAltered, getTotalAltered")
    cqueryopts.add_option("-n", "--study_name", dest="study_name", type="string", default=None,
                          help="[Required/Optional] cBioPortal Name for study [defualt=%default].\n See above for which commands require this.")
    cqueryopts.add_option("-c", "--case_set_id", dest="case_set_id", type="string", default=None,
                          help="[Required for some] cBioPortal case_set_id specifying the case list to use.\nRequired for getProfileData, getMutationData, getClincalData, getProteinArrayData, getPercentAltered, getTotalAltered. Default is case_set_id for case list 'All Tumours' ")
    cqueryopts.add_option("-g", "--gene_list", dest="gene_list", type="string", default=None,
                          help="[Required for some] Comma seperated list of HUGO gene symbols or Entrez gene IDs.\nRequired for getProfileData, getMutationData, getLink, getOncoprintHTML")
    cqueryopts.add_option("-f", "--gene_list_file", dest="gene_list_file", type="string", default=None,
                          help="[Optional] Filename to read in gene_list from")
    cqueryopts.add_option("-p", "--profile_id", dest="profile_id", type="string",
                          help="[Optional] Comma seperated list of cBioPortal genetic_profile_ids. If none are specified then the list of profiles for the study where display in analysis is True is used.")

    squeryopts = optparse.OptionGroup(
        parser, "Query specific parameters", "Arguments specific to a particular query")
    squeryopts.add_option("--protein_array_type", dest="protein_array_type", type="string", default="protein_level",
                          help="[Optional] Either protein_level or phosphorylation [default=%default]")
    squeryopts.add_option("--protein_array_id", dest="protein_array_id", type="string",
                          help="[Required for some] comma seperated list of one or more protein array IDs")
    squeryopts.add_option("--array_info", dest="protein_array_info", type="int",  default=0,
                          help="[Optional] If 1, antibody infomation will also be exported in a getProteinArrayData query [default=%default]")
    squeryopts.add_option("--output-report", dest="report", type="string", default="full",
                          help="[Optional] Report type to display for getLink. Either full or oncoprint_html [default=%default] ")
    squeryopts.add_option("--threshold", dest="threshold", type="int", default=2,
                          help="[Optional] Threshold for deciding if an alteration is significant for continuous metrics [default=%default]")

    parser.add_option_group(cqueryopts)
    parser.add_option_group(squeryopts)

    (options, args) = E.Start(
        parser, add_pipe_options=False, add_output_options=False, argv=argv)

    portal = CBioPortal(url=options.url, study=options.study_id,
                        study_name=options.study_name, case_list_id=options.case_set_id)

    results = []

    if options.gene_list_file:
        infile = IOTools.openFile(options.gene_list_file)
        gene_list = [x.strip() for x in infile]
    elif options.gene_list:
        gene_list = options.gene_list.split(",")

    if options.profile_id:
        profile_id = options.profile_id.split(",")
    else:
        profile_id = None

    if "getCancerStudies" in args:
        results.append(portal.getCancerStudies())

    if "getGeneticProfiles" in args:
        results.append(portal.getGeneticProfiles())

    if "getCaseLists" in args:
        results.append(portal.getCaseLists())

    if "getProfileData" in args:
        results.append(
            portal.getProfileData(gene_list=gene_list,
                                  genetic_profile_id=profile_id))

    if "getMutationData" in args:
        results.append(
            portal.getMutationData(gene_list=gene_list,
                                   genetic_profile_id=profile_id))

    if "getClinicalData" in args:
        results.append(portal.getClinicalData())

    if "getProteinArrayInfo" in args:
        results.append(portal.getProteinArrayInfo(
            gene_list=gene_list,
            protein_array_type=options.protein_array_type))

    if "getProteinArrayData" in args:
        results.append(portal.getProteinArrayData(
            protein_array_id=options.protein_array_id,
            array_info=options.array_info))

    if "getPercentAltered" in args:
        results.append(portal.getPercentAltered(
            gene_list=gene_list, genetic_profile_id=profile_id,
            threshold=options.threshold))

    if "getLink" in args:
        results.append(
            portal.getLink(gene_list=gene_list, report=options.report))

    if "getOncoprintHTML" in args:
        results.append(portal.getOncoprintHTML(gene_list=gene_list))

    if len(results) == 0:
        sys.stderr.write("No recognised query commands provided")
        sys.exit()

    if options.output_file:
        outf = IOTools.openFile(options.output_file, "w")
    else:
        outf = sys.stdout

    for result in results:
        try:
            outf.write(tableToString(result))
        except:
            outf.write(result)

    E.Stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
fastas2fasta.py - concatenate sequences from multiple fasta files
==========================================================================

:Tags: Genomics Sequences MultipleAlignments FASTA Manipulation

Purpose
-------

This script reads sequences from two or more :term:`fasta` formatted
files and outputs a new file with the sequences concatenated per
entry.

All files must have the same number of sequences and the id of
the first file is output.

Usage
-----

Example::

   python fastas2fasta.py a.fasta b.fasta > c.fasta

If a.fasta is::

  >1
  AAACC
  >2
  CCCAA

and b.fasta is::

  >a
  GGGGTTT
  >b
  TTTTGGG

then the output will be::

  >1
  AAACCGGGGTTT
  >2
  CCCAATTTTGGG


Type::

   python fastas2fasta.py --help

for command line help.

Command line options
--------------------

'''
import sys
import re

import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.FastaIterator as FastaIterator


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(version="%prog version: $Id: fastas2fasta.py 2782 2009-09-10 11:40:29Z andreas $",
                            usage=globals()["__doc__"])

    (options, args) = E.Start(parser)

    if len(args) < 2:
        raise ValueError(
            "please supply at least two filenames to concatenate.")

    iterators = []
    for a in args:
        iterators.append(FastaIterator.FastaIterator(IOTools.openFile(a, "r")))

    ninput, noutput, nerrors = 0, 0, 0

    while 1:

        sequences = []
        ids = []

        for iterator in iterators:
            try:
                cur_record = next(iterator)
            except StopIteration:
                break

            sequences.append(re.sub(" ", "", cur_record.sequence))
            ids.append(cur_record.title)

        if not sequences:
            break
        ninput += 1

        if len(sequences) != len(iterators):
            raise ValueError("unequal number of sequences in files")

        noutput += 1

        options.stdout.write(">%s\n%s\n" % (ids[0],
                                            "".join(sequences)))

    E.info("ninput=%i, noutput=%i, nerrors=%i" % (ninput, noutput, nerrors))

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))
